"""SSH-based Kubernetes Cluster Deployment Script"""
# pylint: disable=line-too-long
import base64
import concurrent.futures as cf
import os
import re
import shlex
import shutil
import tempfile
from typing import List, Optional

import colorama
import yaml

from sky import sky_logging
from sky.ssh_node_pools import constants
from sky.ssh_node_pools import utils as ssh_utils
from sky.ssh_node_pools.deploy import tunnel_utils
from sky.ssh_node_pools.deploy import utils as deploy_utils
from sky.utils import rich_utils
from sky.utils import ux_utils

RESET_ALL = colorama.Style.RESET_ALL

# Get the directory of this script
SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))

logger = sky_logging.init_logger(__name__)


def progress_message(message):
    """Show a progress message."""
    logger.info(f'{colorama.Fore.YELLOW}➜ {message}{RESET_ALL}')


def success_message(message):
    """Show a success message."""
    logger.info(f'{colorama.Fore.GREEN}✔ {message}{RESET_ALL}')


def force_update_status(message):
    """Force update rich spinner status."""
    rich_utils.force_update_status(ux_utils.spinner_message(message))


def run(cleanup: bool = False,
        infra: Optional[str] = None,
        kubeconfig_path: str = constants.DEFAULT_KUBECONFIG_PATH):
    """Deploy a Kubernetes cluster on SSH targets.

    This function reads ~/.sky/ssh_node_pools.yaml and uses it to deploy a
    Kubernetes cluster on the specified machines.

    Args:
        cleanup: Whether to clean up the cluster instead of deploying.
        infra: Name of the cluster in ssh_node_pools.yaml to use.
            If None, the first cluster in the file will be used.
        kubeconfig_path: Path to save the Kubernetes configuration file.
            If None, the default ~/.kube/config will be used.
    """
    deploy_utils.check_ssh_cluster_dependencies()
    action = 'Cleanup' if cleanup else 'Deployment'
    msg_str = f'Initializing SSH Node Pools {action}...'

    with rich_utils.safe_status(ux_utils.spinner_message(msg_str)):
        try:
            deploy_multiple_clusters(infra=infra,
                                     cleanup=cleanup,
                                     kubeconfig_path=kubeconfig_path)
        except Exception as e:  # pylint: disable=broad-except
            logger.error(str(e))
            with ux_utils.print_exception_no_traceback():
                raise RuntimeError(
                    'Failed to deploy SkyPilot on some Node Pools.') from e

    # Add empty line for ux-purposes.
    logger.info('')
    if cleanup:
        logger.info(
            ux_utils.finishing_message(
                '🎉 SSH Node Pools cleaned up successfully.'))
    else:
        logger.info(
            ux_utils.finishing_message(
                '🎉 SSH Node Pools set up successfully. ',
                follow_up_message=(
                    f'Run `{colorama.Style.BRIGHT}'
                    f'sky check ssh'
                    f'{colorama.Style.RESET_ALL}` to verify access, '
                    f'`{colorama.Style.BRIGHT}sky launch --infra ssh'
                    f'{colorama.Style.RESET_ALL}` to launch a cluster.')))


def deploy_multiple_clusters(
        infra: Optional[str],
        ssh_node_pools_file: str = constants.DEFAULT_SSH_NODE_POOLS_PATH,
        kubeconfig_path: str = constants.DEFAULT_KUBECONFIG_PATH,
        cleanup: bool = True):

    kubeconfig_path = kubeconfig_path or constants.DEFAULT_KUBECONFIG_PATH
    kubeconfig_path = os.path.expanduser(kubeconfig_path)

    failed_clusters = []
    successful_clusters = []

    # Using YAML configuration
    targets = ssh_utils.load_ssh_targets(ssh_node_pools_file)
    clusters_config = ssh_utils.get_cluster_config(
        targets, infra, file_path=ssh_node_pools_file)

    # Print information about clusters being processed
    num_clusters = len(clusters_config)
    cluster_names = list(clusters_config.keys())
    cluster_info = f'Found {num_clusters} Node Pool{"s" if num_clusters > 1 else ""}: {", ".join(cluster_names)}'
    logger.info(f'{colorama.Fore.CYAN}{cluster_info}{RESET_ALL}')

    # Process each cluster
    for cluster_name, cluster_config in clusters_config.items():
        try:
            action = 'Cleaning up' if cleanup else 'Deploying'
            force_update_status(f'{action} Node Pool: {cluster_name}')
            hosts_info = ssh_utils.prepare_hosts_info(cluster_name,
                                                      cluster_config)

            if not hosts_info:
                logger.warning(
                    f'{colorama.Fore.RED}Error: No valid hosts found '
                    f'for cluster {cluster_name!r}. Skipping.{RESET_ALL}')
                continue

            context_name = f'ssh-{cluster_name}'

            # Check cluster history
            os.makedirs(constants.NODE_POOLS_INFO_DIR, exist_ok=True)
            history_yaml_file = os.path.join(constants.NODE_POOLS_INFO_DIR,
                                             f'{context_name}-history.yaml')

            history = None
            if os.path.exists(history_yaml_file):
                logger.debug(f'Loading history from {history_yaml_file}')
                with open(history_yaml_file, 'r', encoding='utf-8') as f:
                    history = yaml.safe_load(f)
            else:
                logger.debug(f'No history found for {context_name}.')

            history_workers_info = None
            history_worker_nodes = None
            history_use_ssh_config = None
            # Do not support changing anything besides hosts for now
            if history is not None:
                for key in ['user', 'identity_file', 'password']:
                    if not cleanup and history.get(key) != cluster_config.get(
                            key):
                        raise ValueError(
                            f'Cluster configuration has changed for field {key!r}. '
                            f'Previous value: {history.get(key)}, '
                            f'Current value: {cluster_config.get(key)}')
                history_hosts_info = ssh_utils.prepare_hosts_info(
                    cluster_name, history)
                if not cleanup and history_hosts_info[0] != hosts_info[0]:
                    raise ValueError(
                        f'Cluster configuration has changed for master node. '
                        f'Previous value: {history_hosts_info[0]}, '
                        f'Current value: {hosts_info[0]}')
                history_workers_info = history_hosts_info[1:] if len(
                    history_hosts_info) > 1 else []
                history_worker_nodes = [h['ip'] for h in history_workers_info]
                history_use_ssh_config = [
                    h.get('use_ssh_config', False) for h in history_workers_info
                ]

            # Use the first host as the head node and the rest as worker nodes
            head_host = hosts_info[0]
            worker_hosts = hosts_info[1:] if len(hosts_info) > 1 else []

            head_node = head_host['ip']
            worker_nodes = [h['ip'] for h in worker_hosts]
            ssh_user = head_host['user']
            ssh_key = head_host['identity_file']
            head_use_ssh_config = head_host.get('use_ssh_config', False)
            worker_use_ssh_config = [
                h.get('use_ssh_config', False) for h in worker_hosts
            ]
            password = head_host['password']

            # Deploy this cluster
            unsuccessful_workers = deploy_single_cluster(
                cluster_name,
                head_node,
                worker_nodes,
                ssh_user,
                ssh_key,
                context_name,
                password,
                head_use_ssh_config,
                worker_use_ssh_config,
                kubeconfig_path,
                cleanup,
                worker_hosts=worker_hosts,
                history_worker_nodes=history_worker_nodes,
                history_workers_info=history_workers_info,
                history_use_ssh_config=history_use_ssh_config)

            if not cleanup:
                successful_hosts = []
                for host in cluster_config['hosts']:
                    if isinstance(host, str):
                        host_node = host
                    else:
                        host_node = host['ip']
                    if host_node not in unsuccessful_workers:
                        successful_hosts.append(host)
                cluster_config['hosts'] = successful_hosts
                with open(history_yaml_file, 'w', encoding='utf-8') as f:
                    logger.debug(f'Writing history to {history_yaml_file}')
                    yaml.dump(cluster_config, f)

            action = 'cleanup' if cleanup else 'deployment'
            logger.info(
                f'{colorama.Fore.CYAN}Completed {action} for cluster: {cluster_name}{colorama.Style.RESET_ALL}'
            )
            successful_clusters.append(cluster_name)
        except Exception as e:  # pylint: disable=broad-except
            reason = str(e)
            failed_clusters.append((cluster_name, reason))
            action = 'cleaning' if cleanup else 'deploying'
            logger.debug(
                f'Error {action} SSH Node Pool `{cluster_name}`: {reason}')

    if failed_clusters:
        action = 'clean' if cleanup else 'deploy'
        msg = f'{colorama.Fore.GREEN}Successfully {action}ed {len(successful_clusters)} cluster(s) ({", ".join(successful_clusters)}). {RESET_ALL}'
        msg += f'{colorama.Fore.RED}Failed to {action} {len(failed_clusters)} cluster(s): {RESET_ALL}'
        for cluster_name, reason in failed_clusters:
            msg += f'\n  {cluster_name}: {reason}'
        raise RuntimeError(msg)


def deploy_single_cluster(cluster_name,
                          head_node,
                          worker_nodes,
                          ssh_user,
                          ssh_key,
                          context_name,
                          password,
                          head_use_ssh_config,
                          worker_use_ssh_config,
                          kubeconfig_path,
                          cleanup,
                          worker_hosts=None,
                          history_worker_nodes=None,
                          history_workers_info=None,
                          history_use_ssh_config=None) -> List[str]:
    """Deploy or clean up a single Kubernetes cluster.

    Returns: List of unsuccessful worker nodes.
    """
    history_yaml_file = os.path.join(constants.NODE_POOLS_INFO_DIR,
                                     f'{context_name}-history.yaml')
    cert_file_path = os.path.join(constants.NODE_POOLS_INFO_DIR,
                                  f'{context_name}-cert.pem')
    key_file_path = os.path.join(constants.NODE_POOLS_INFO_DIR,
                                 f'{context_name}-key.pem')
    tunnel_log_file_path = os.path.join(constants.NODE_POOLS_INFO_DIR,
                                        f'{context_name}-tunnel.log')

    # Generate the askpass block if password is provided
    askpass_block = create_askpass_script(password)

    # Token for k3s
    # TODO (kyuds): make this configurable?
    k3s_token = constants.K3S_TOKEN

    # Pre-flight checks
    logger.info(f'Checking SSH connection to head node ({head_node})...')
    result = deploy_utils.run_remote(
        head_node,
        f'echo \'SSH connection successful ({head_node})\'',
        ssh_user,
        ssh_key,
        use_ssh_config=head_use_ssh_config)
    if result is None:
        with ux_utils.print_exception_no_traceback():
            raise RuntimeError(
                f'Failed to SSH to head node ({head_node}). '
                f'Please check the SSH configuration and logs for more details.'
            )
    elif result.startswith('SSH connection successful'):
        success_message(f'SSH connection established to head node {head_node}.')

    # Checking history
    history_exists = (history_worker_nodes is not None and
                      history_workers_info is not None and
                      history_use_ssh_config is not None)

    # Cleanup history worker nodes
    worker_nodes_to_cleanup = []
    remove_worker_cmds = []
    if history_exists:
        for history_node, history_info, use_ssh_config in zip(
                history_worker_nodes, history_workers_info,
                history_use_ssh_config):
            if worker_hosts is not None and history_info not in worker_hosts:
                logger.debug(
                    f'Worker node {history_node} not found in YAML config. '
                    'Removing from history...')
                worker_nodes_to_cleanup.append(
                    dict(
                        node=history_node,
                        user=ssh_user
                        if history_info is None else history_info['user'],
                        ssh_key=ssh_key if history_info is None else
                        history_info['identity_file'],
                        askpass_block=(askpass_block if history_info is None
                                       else create_askpass_script(
                                           history_info['password'])),
                        use_ssh_config=use_ssh_config,
                    ))
                remove_worker_cmds.append(
                    f'kubectl delete node -l skypilot-ip={history_node}')
        # If this is a create operation and there exists some stale log,
        # cleanup the log for a new file to store new logs.
        if not cleanup and os.path.exists(tunnel_log_file_path):
            os.remove(tunnel_log_file_path)

    # If --cleanup flag is set, uninstall k3s and exit
    if cleanup:
        # Pickup all nodes
        worker_nodes_to_cleanup.clear()
        for node, info, use_ssh_config in zip(worker_nodes, worker_hosts,
                                              worker_use_ssh_config):
            worker_nodes_to_cleanup.append(
                dict(
                    node=node,
                    user=ssh_user if info is None else info['user'],
                    ssh_key=ssh_key if info is None else info['identity_file'],
                    askpass_block=(askpass_block if info is None else
                                   create_askpass_script(info['password'])),
                    use_ssh_config=use_ssh_config,
                ))

        # Clean up head node
        cleanup_node(head_node,
                     ssh_user,
                     ssh_key,
                     askpass_block,
                     use_ssh_config=head_use_ssh_config,
                     is_worker=False)
    # Clean up worker nodes
    force_update_status(f'Cleaning up worker nodes [{cluster_name}]')
    with cf.ThreadPoolExecutor() as executor:
        executor.map(lambda kwargs: cleanup_node(**kwargs),
                     worker_nodes_to_cleanup)

    with cf.ThreadPoolExecutor() as executor:
        executor.map(lambda cmd: deploy_utils.run_command(cmd, shell=True),
                     remove_worker_cmds)

    if cleanup:
        # Remove the context from local kubeconfig if it exists
        if os.path.isfile(kubeconfig_path):
            logger.debug(
                f'Removing context {context_name!r} from local kubeconfig...')
            deploy_utils.run_command(
                ['kubectl', 'config', 'delete-context', context_name],
                shell=False,
                silent=True)
            deploy_utils.run_command(
                ['kubectl', 'config', 'delete-cluster', context_name],
                shell=False,
                silent=True)
            deploy_utils.run_command(
                ['kubectl', 'config', 'delete-user', context_name],
                shell=False,
                silent=True)

            # Update the current context to the first available context
            contexts = deploy_utils.run_command([
                'kubectl', 'config', 'view', '-o',
                'jsonpath=\'{.contexts[0].name}\''
            ],
                                                shell=False,
                                                silent=True)
            if contexts:
                deploy_utils.run_command(
                    ['kubectl', 'config', 'use-context', contexts],
                    shell=False,
                    silent=True)
            else:
                # If no context is available, simply unset the current context
                deploy_utils.run_command(
                    ['kubectl', 'config', 'unset', 'current-context'],
                    shell=False,
                    silent=True)

            logger.debug(
                f'Context {context_name!r} removed from local kubeconfig.')

        for file in [history_yaml_file, cert_file_path, key_file_path]:
            if os.path.exists(file):
                os.remove(file)

        # Clean up SSH tunnel after clean up kubeconfig, because the kubectl
        # will restart the ssh tunnel if it's not running.
        tunnel_utils.cleanup_kubectl_ssh_tunnel(cluster_name, context_name)

        success_message(f'Node Pool `{cluster_name}` cleaned up successfully.')
        return []

    logger.debug('Checking TCP Forwarding Options...')
    cmd = (
        'if [ "$(sudo sshd -T | grep allowtcpforwarding)" = "allowtcpforwarding yes" ]; then '
        f'echo "TCP Forwarding already enabled on head node ({head_node})."; '
        'else '
        'sudo sed -i \'s/^#\\?\\s*AllowTcpForwarding.*/AllowTcpForwarding yes/\' '
        '/etc/ssh/sshd_config && sudo systemctl restart sshd && '
        f'echo "Successfully enabled TCP Forwarding on head node ({head_node})."; '
        'fi')
    result = deploy_utils.run_remote(head_node,
                                     shlex.quote(cmd),
                                     ssh_user,
                                     ssh_key,
                                     use_ssh_config=head_use_ssh_config,
                                     use_shell=True)
    if result is None:
        with ux_utils.print_exception_no_traceback():
            raise RuntimeError(
                f'Failed to setup TCP forwarding on head node ({head_node}). '
                f'Please check the SSH configuration.')

    # Get effective IP for master node if using SSH config - needed for workers to connect
    if head_use_ssh_config:
        effective_master_ip = deploy_utils.get_effective_host_ip(head_node)
        logger.info(f'{colorama.Fore.GREEN}Resolved head node {head_node} '
                    f'to {effective_master_ip} from SSH config{RESET_ALL}')
    else:
        effective_master_ip = head_node

    # Step 1: Install k3s on the head node
    # Check if head node has a GPU
    install_gpu = False
    force_update_status(
        f'Deploying SkyPilot runtime on head node ({head_node}).')
    cmd = f"""
        {askpass_block}
        curl -sfL https://get.k3s.io | K3S_TOKEN={k3s_token} K3S_NODE_NAME={head_node} sudo -E -A sh - &&
        mkdir -p ~/.kube &&
        sudo -A cp /etc/rancher/k3s/k3s.yaml ~/.kube/config &&
        sudo -A chown $(id -u):$(id -g) ~/.kube/config &&
        for i in {{1..3}}; do
            if kubectl wait --for=condition=ready node --all --timeout=2m --kubeconfig ~/.kube/config; then
                break
            else
                echo 'Waiting for nodes to be ready...'
                sleep 5
            fi
        done
        if [ $i -eq 3 ]; then
            echo 'Failed to wait for nodes to be ready after 3 attempts'
            exit 1
        fi
    """
    result = deploy_utils.run_remote(head_node,
                                     cmd,
                                     ssh_user,
                                     ssh_key,
                                     use_ssh_config=head_use_ssh_config)
    if result is None:
        with ux_utils.print_exception_no_traceback():
            raise RuntimeError(
                f'Failed to deploy K3s on head node ({head_node}).')
    success_message(
        f'SkyPilot runtime successfully deployed on head node ({head_node}).')

    # Check if head node has a GPU
    install_gpu = False
    if deploy_utils.check_gpu(head_node,
                              ssh_user,
                              ssh_key,
                              use_ssh_config=head_use_ssh_config,
                              is_head=True):
        install_gpu = True

    # Fetch the head node's internal IP (this will be passed to worker nodes)
    master_addr = deploy_utils.run_remote(head_node,
                                          'hostname -I | awk \'{print $1}\'',
                                          ssh_user,
                                          ssh_key,
                                          use_ssh_config=head_use_ssh_config)
    if master_addr is None:
        with ux_utils.print_exception_no_traceback():
            raise RuntimeError(f'Failed to SSH to head node ({head_node}). '
                               f'Please check the SSH configuration.')
    logger.debug(f'Master node internal IP: {master_addr}')

    # Step 2: Install k3s on worker nodes and join them to the master node
    def deploy_worker(args):
        (i, node, worker_hosts, history_workers_info, ssh_user, ssh_key,
         askpass_block, worker_use_ssh_config, master_addr, k3s_token) = args

        # If using YAML config with specific worker info
        if worker_hosts and i < len(worker_hosts):
            if history_workers_info is not None and worker_hosts[
                    i] in history_workers_info:
                logger.info(
                    f'{colorama.Style.DIM}✔ SkyPilot runtime already deployed on worker node {node}. '
                    f'Skipping...{RESET_ALL}')
                return node, True, False
            worker_user = worker_hosts[i]['user']
            worker_key = worker_hosts[i]['identity_file']
            worker_password = worker_hosts[i]['password']
            worker_askpass = create_askpass_script(worker_password)
            worker_config = worker_use_ssh_config[i]
        else:
            worker_user = ssh_user
            worker_key = ssh_key
            worker_askpass = askpass_block
            worker_config = worker_use_ssh_config[i]

        return start_agent_node(node,
                                master_addr,
                                k3s_token,
                                worker_user,
                                worker_key,
                                worker_askpass,
                                use_ssh_config=worker_config)

    unsuccessful_workers = []

    # Deploy workers in parallel using thread pool
    force_update_status(
        f'Deploying SkyPilot runtime on worker nodes [{cluster_name}]')
    with cf.ThreadPoolExecutor() as executor:
        futures = []
        for i, node in enumerate(worker_nodes):
            args = (i, node, worker_hosts, history_workers_info, ssh_user,
                    ssh_key, askpass_block, worker_use_ssh_config, master_addr,
                    k3s_token)
            futures.append(executor.submit(deploy_worker, args))

        # Check if worker node has a GPU
        for future in cf.as_completed(futures):
            node, suc, has_gpu = future.result()
            install_gpu = install_gpu or has_gpu
            if not suc:
                unsuccessful_workers.append(node)

    # Step 3: Configure local kubectl to connect to the cluster
    force_update_status(f'Setting up SkyPilot configuration [{cluster_name}]')

    # Create temporary directory for kubeconfig operations
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_kubeconfig = os.path.join(temp_dir, 'kubeconfig')

        # Get the kubeconfig from remote server
        if head_use_ssh_config:
            scp_cmd = ['scp', head_node + ':~/.kube/config', temp_kubeconfig]
        else:
            scp_cmd = [
                'scp', '-o', 'StrictHostKeyChecking=no', '-o',
                'IdentitiesOnly=yes', '-i', ssh_key,
                f'{ssh_user}@{head_node}:~/.kube/config', temp_kubeconfig
            ]
        deploy_utils.run_command(scp_cmd, shell=False)

        # Create the directory for the kubeconfig file if it doesn't exist
        deploy_utils.ensure_directory_exists(kubeconfig_path)

        # Create empty kubeconfig if it doesn't exist
        if not os.path.isfile(kubeconfig_path):
            open(kubeconfig_path, 'a', encoding='utf-8').close()

        # Modify the temporary kubeconfig to update server address and context name
        modified_config = os.path.join(temp_dir, 'modified_config')
        with open(temp_kubeconfig, 'r', encoding='utf-8') as f_in:
            with open(modified_config, 'w', encoding='utf-8') as f_out:
                in_cluster = False
                in_user = False
                client_cert_data = None
                client_key_data = None

                for line in f_in:
                    if 'clusters:' in line:
                        in_cluster = True
                        in_user = False
                    elif 'users:' in line:
                        in_cluster = False
                        in_user = True
                    elif 'contexts:' in line:
                        in_cluster = False
                        in_user = False

                    # Skip certificate authority data in cluster section
                    if in_cluster and 'certificate-authority-data:' in line:
                        continue
                    # Skip client certificate data in user section but extract it
                    elif in_user and 'client-certificate-data:' in line:
                        client_cert_data = line.split(':', 1)[1].strip()
                        continue
                    # Skip client key data in user section but extract it
                    elif in_user and 'client-key-data:' in line:
                        client_key_data = line.split(':', 1)[1].strip()
                        continue
                    elif in_cluster and 'server:' in line:
                        # Initially just set to the effective master IP
                        # (will be changed to localhost by setup_kubectl_ssh_tunnel later)
                        f_out.write(
                            f'    server: https://{effective_master_ip}:6443\n')
                        f_out.write('    insecure-skip-tls-verify: true\n')
                        continue

                    # Replace default context names with user-provided context name
                    line = line.replace('name: default',
                                        f'name: {context_name}')
                    line = line.replace('cluster: default',
                                        f'cluster: {context_name}')
                    line = line.replace('user: default',
                                        f'user: {context_name}')
                    line = line.replace('current-context: default',
                                        f'current-context: {context_name}')

                    f_out.write(line)

                # Save certificate data if available

                if client_cert_data:
                    # Decode base64 data and save as PEM
                    try:
                        # Clean up the certificate data by removing whitespace
                        clean_cert_data = ''.join(client_cert_data.split())
                        cert_pem = base64.b64decode(clean_cert_data).decode(
                            'utf-8')

                        # Check if the data already looks like a PEM file
                        has_begin = '-----BEGIN CERTIFICATE-----' in cert_pem
                        has_end = '-----END CERTIFICATE-----' in cert_pem

                        if not has_begin or not has_end:
                            logger.debug(
                                'Warning: Certificate data missing PEM markers, attempting to fix...'
                            )
                            # Add PEM markers if missing
                            if not has_begin:
                                cert_pem = f'-----BEGIN CERTIFICATE-----\n{cert_pem}'
                            if not has_end:
                                cert_pem = f'{cert_pem}\n-----END CERTIFICATE-----'

                        # Write the certificate
                        with open(cert_file_path, 'w',
                                  encoding='utf-8') as cert_file:
                            cert_file.write(cert_pem)

                        # Verify the file was written correctly
                        if os.path.getsize(cert_file_path) > 0:
                            logger.debug(
                                f'Successfully saved certificate data ({len(cert_pem)} bytes)'
                            )

                            # Quick validation of PEM format
                            with open(cert_file_path, 'r',
                                      encoding='utf-8') as f:
                                content = f.readlines()
                                first_line = content[0].strip(
                                ) if content else ''
                                last_line = content[-1].strip(
                                ) if content else ''

                            if not first_line.startswith(
                                    '-----BEGIN') or not last_line.startswith(
                                        '-----END'):
                                logger.debug(
                                    'Warning: Certificate may not be in proper PEM format'
                                )
                        else:
                            logger.error(
                                f'{colorama.Fore.RED}Error: '
                                f'Certificate file is empty{RESET_ALL}')
                    except Exception as e:  # pylint: disable=broad-except
                        logger.error(f'{colorama.Fore.RED}'
                                     f'Error processing certificate data: {e}'
                                     f'{RESET_ALL}')

                if client_key_data:
                    # Decode base64 data and save as PEM
                    try:
                        # Clean up the key data by removing whitespace
                        clean_key_data = ''.join(client_key_data.split())
                        key_pem = base64.b64decode(clean_key_data).decode(
                            'utf-8')

                        # Check if the data already looks like a PEM file

                        # Check for EC key format
                        if 'EC PRIVATE KEY' in key_pem:
                            # Handle EC KEY format directly
                            match_ec = re.search(
                                r'-----BEGIN EC PRIVATE KEY-----(.*?)-----END EC PRIVATE KEY-----',
                                key_pem, re.DOTALL)
                            if match_ec:
                                # Extract and properly format EC key
                                key_content = match_ec.group(1).strip()
                                key_pem = f'-----BEGIN EC PRIVATE KEY-----\n{key_content}\n-----END EC PRIVATE KEY-----'
                            else:
                                # Extract content and assume EC format
                                key_content = re.sub(r'-----BEGIN.*?-----', '',
                                                     key_pem)
                                key_content = re.sub(r'-----END.*?-----.*', '',
                                                     key_content).strip()
                                key_pem = f'-----BEGIN EC PRIVATE KEY-----\n{key_content}\n-----END EC PRIVATE KEY-----'
                        else:
                            # Handle regular private key format
                            has_begin = any(marker in key_pem for marker in [
                                '-----BEGIN PRIVATE KEY-----',
                                '-----BEGIN RSA PRIVATE KEY-----'
                            ])
                            has_end = any(marker in key_pem for marker in [
                                '-----END PRIVATE KEY-----',
                                '-----END RSA PRIVATE KEY-----'
                            ])

                            if not has_begin or not has_end:
                                logger.debug(
                                    'Warning: Key data missing PEM markers, attempting to fix...'
                                )
                                # Add PEM markers if missing
                                if not has_begin:
                                    key_pem = f'-----BEGIN PRIVATE KEY-----\n{key_pem}'
                                if not has_end:
                                    key_pem = f'{key_pem}\n-----END PRIVATE KEY-----'
                                    # Remove any trailing characters after END marker
                                    key_pem = re.sub(
                                        r'(-----END PRIVATE KEY-----).*', r'\1',
                                        key_pem)

                        # Write the key
                        with open(key_file_path, 'w',
                                  encoding='utf-8') as key_file:
                            key_file.write(key_pem)

                        # Verify the file was written correctly
                        if os.path.getsize(key_file_path) > 0:
                            logger.debug(
                                f'Successfully saved key data ({len(key_pem)} bytes)'
                            )

                            # Quick validation of PEM format
                            with open(key_file_path, 'r',
                                      encoding='utf-8') as f:
                                content = f.readlines()
                                first_line = content[0].strip(
                                ) if content else ''
                                last_line = content[-1].strip(
                                ) if content else ''

                            if not first_line.startswith(
                                    '-----BEGIN') or not last_line.startswith(
                                        '-----END'):
                                logger.debug(
                                    'Warning: Key may not be in proper PEM format'
                                )
                        else:
                            logger.error(f'{colorama.Fore.RED}Error: '
                                         f'Key file is empty{RESET_ALL}')
                    except Exception as e:  # pylint: disable=broad-except
                        logger.error(f'{colorama.Fore.RED}'
                                     f'Error processing key data: {e}'
                                     f'{RESET_ALL}')

        # First check if context name exists and delete it if it does
        # TODO(romilb): Should we throw an error here instead?
        deploy_utils.run_command(
            ['kubectl', 'config', 'delete-context', context_name],
            shell=False,
            silent=True)
        deploy_utils.run_command(
            ['kubectl', 'config', 'delete-cluster', context_name],
            shell=False,
            silent=True)
        deploy_utils.run_command(
            ['kubectl', 'config', 'delete-user', context_name],
            shell=False,
            silent=True)

        # Merge the configurations using kubectl
        merged_config = os.path.join(temp_dir, 'merged_config')
        os.environ['KUBECONFIG'] = f'{kubeconfig_path}:{modified_config}'
        with open(merged_config, 'w', encoding='utf-8') as merged_file:
            kubectl_cmd = ['kubectl', 'config', 'view', '--flatten']
            result = deploy_utils.run_command(kubectl_cmd, shell=False)
            if result:
                merged_file.write(result)

        # Replace the kubeconfig with the merged config
        shutil.move(merged_config, kubeconfig_path)

        # Set the new context as the current context
        deploy_utils.run_command(
            ['kubectl', 'config', 'use-context', context_name],
            shell=False,
            silent=True)

    # Always set up SSH tunnel since we assume only port 22 is accessible
    tunnel_utils.setup_kubectl_ssh_tunnel(head_node,
                                          ssh_user,
                                          ssh_key,
                                          context_name,
                                          use_ssh_config=head_use_ssh_config)

    logger.debug(f'kubectl configured with new context \'{context_name}\'.')
    success_message(f'SkyPilot runtime is up [{cluster_name}].')

    # Install GPU operator if a GPU was detected on any node
    if install_gpu:
        force_update_status(f'Configuring NVIDIA GPUs [{cluster_name}]')
        cmd = f"""
            {askpass_block}
            curl -fsSL -o get_helm.sh https://raw.githubusercontent.com/helm/helm/master/scripts/get-helm-3 &&
            chmod 700 get_helm.sh &&
            ./get_helm.sh &&
            helm repo add nvidia https://helm.ngc.nvidia.com/nvidia && helm repo update &&
            kubectl create namespace gpu-operator --kubeconfig ~/.kube/config || true &&
            sudo -A ln -s /sbin/ldconfig /sbin/ldconfig.real || true &&
            helm install gpu-operator -n gpu-operator --create-namespace nvidia/gpu-operator \\
            --set 'toolkit.env[0].name=CONTAINERD_CONFIG' \\
            --set 'toolkit.env[0].value=/var/lib/rancher/k3s/agent/etc/containerd/config.toml' \\
            --set 'toolkit.env[1].name=CONTAINERD_SOCKET' \\
            --set 'toolkit.env[1].value=/run/k3s/containerd/containerd.sock' \\
            --set 'toolkit.env[2].name=CONTAINERD_RUNTIME_CLASS' \\
            --set 'toolkit.env[2].value=nvidia' &&
            echo 'Waiting for GPU operator installation...' &&
            while ! kubectl describe nodes --kubeconfig ~/.kube/config | grep -q 'nvidia.com/gpu:' || ! kubectl describe nodes --kubeconfig ~/.kube/config | grep -q 'nvidia.com/gpu.product'; do
                echo 'Waiting for GPU operator...'
                sleep 5
            done
            echo 'GPU operator installed successfully.'
        """
        result = deploy_utils.run_remote(head_node,
                                         cmd,
                                         ssh_user,
                                         ssh_key,
                                         use_ssh_config=head_use_ssh_config)
        if result is None:
            logger.error(f'{colorama.Fore.RED}Failed to install GPU Operator.'
                         f'{RESET_ALL}')
        else:
            success_message('GPU Operator installed.')
    else:
        logger.debug('No GPUs detected. Skipping GPU Operator installation.')

    # The env var KUBECONFIG ensures sky check uses the right kubeconfig
    os.environ['KUBECONFIG'] = kubeconfig_path
    deploy_utils.run_command(['sky', 'check', 'ssh'], shell=False)

    success_message('SkyPilot configured successfully.')

    if unsuccessful_workers:
        quoted_unsuccessful_workers = [
            f'"{worker}"' for worker in unsuccessful_workers
        ]

        logger.info(f'{colorama.Fore.YELLOW}'
                    'Failed to deploy Kubernetes on the following nodes: '
                    f'{", ".join(quoted_unsuccessful_workers)}. Please check '
                    f'the logs for more details.{RESET_ALL}')
    else:
        success_message(f'Node Pool `{cluster_name}` deployed successfully.')

    return unsuccessful_workers


def create_askpass_script(password):
    """Create an askpass script block for sudo with password."""
    if not password:
        return ''

    return f"""
# Create temporary askpass script
ASKPASS_SCRIPT=$(mktemp)
trap 'rm -f $ASKPASS_SCRIPT' EXIT INT TERM ERR QUIT
cat > $ASKPASS_SCRIPT << EOF
#!/bin/bash
echo {password}
EOF
chmod 700 $ASKPASS_SCRIPT
# Use askpass
export SUDO_ASKPASS=$ASKPASS_SCRIPT
"""


def cleanup_node(node,
                 user,
                 ssh_key,
                 askpass_block,
                 use_ssh_config=False,
                 is_worker=True):
    """Uninstall k3s and clean up the state on a node."""
    ntype = 'worker' if is_worker else 'head'
    force_update_status(f'Cleaning up {ntype} node ({node})...')
    script = f'k3s{"-agent" if is_worker else ""}-uninstall.sh'
    cmd = f"""
        {askpass_block}
        echo 'Uninstalling k3s...' &&
        sudo -A /usr/local/bin/{script} || true &&
        sudo -A rm -rf /etc/rancher /var/lib/rancher /var/lib/kubelet /etc/kubernetes ~/.kube
    """
    result = deploy_utils.run_remote(node,
                                     cmd,
                                     user,
                                     ssh_key,
                                     use_ssh_config=use_ssh_config)
    if result is None:
        logger.error(f'{colorama.Fore.RED}Failed to clean up {ntype} '
                     f'node ({node}).{RESET_ALL}')
    else:
        success_message(f'Node {node} cleaned up successfully.')


def start_agent_node(node,
                     master_addr,
                     k3s_token,
                     user,
                     ssh_key,
                     askpass_block,
                     use_ssh_config=False):
    """Start a k3s agent node.
    Returns: if the start is successful, and whether the node has a GPU."""
    logger.info(f'Deploying worker node ({node}).')
    cmd = f"""
            {askpass_block}
            curl -sfL https://get.k3s.io | K3S_NODE_NAME={node} INSTALL_K3S_EXEC='agent --node-label skypilot-ip={node}' \
                K3S_URL=https://{master_addr}:6443 K3S_TOKEN={k3s_token} sudo -E -A sh -
        """
    result = deploy_utils.run_remote(node,
                                     cmd,
                                     user,
                                     ssh_key,
                                     use_ssh_config=use_ssh_config)
    if result is None:
        logger.error(f'{colorama.Fore.RED}✗ Failed to deploy K3s on worker '
                     f'node ({node}).{RESET_ALL}')
        return node, False, False
    success_message(
        f'SkyPilot runtime successfully deployed on worker node ({node}).')
    # Check if worker node has a GPU
    if deploy_utils.check_gpu(node,
                              user,
                              ssh_key,
                              use_ssh_config=use_ssh_config):
        logger.info(f'{colorama.Fore.YELLOW}GPU detected on worker node '
                    f'({node}).{RESET_ALL}')
        return node, True, True
    return node, True, False



"""Utilities for server-side interactive SSH functionality."""
import array
import socket


def get_pty_socket_path(session_id: str) -> str:
    """Get the Unix socket path for PTY file descriptor passing."""
    return f'/tmp/sky_pty_{session_id}.sock'


def send_fd(sock: socket.socket, fd: int) -> None:
    """Send file descriptor via Unix socket using SCM_RIGHTS.

    SCM_RIGHTS allows us to send or receive a set of open
    file descriptors from another process.

    See:
    https://man7.org/linux/man-pages/man7/unix.7.html
    https://man7.org/linux/man-pages/man3/cmsg.3.html

    Args:
        sock: Connected Unix socket.
        fd: File descriptor to send.
    """
    sock.sendmsg(
        [b'x'],  # Dummy data
        [(socket.SOL_SOCKET, socket.SCM_RIGHTS, array.array('i', [fd]))])


def recv_fd(sock: socket.socket) -> int:
    """Receive file descriptor via Unix socket using SCM_RIGHTS.

    Args:
        sock: Connected Unix socket.

    Returns:
        Received file descriptor.

    Raises:
        RuntimeError: If no file descriptor was received.
    """
    # NOTE: recvmsg() has no async equivalent
    _, ancdata, _, _ = sock.recvmsg(
        1, socket.CMSG_SPACE(array.array('i', [0]).itemsize))
    if not ancdata:
        raise RuntimeError('No file descriptor received - '
                           'sender may have closed connection')
    _, _, cmsg_data = ancdata[0]
    return array.array('i', cmsg_data)[0]



"""Vast cloud adaptor."""

import functools

_vast_sdk = None


def import_package(func):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        global _vast_sdk

        if _vast_sdk is None:
            try:
                import vastai_sdk as _vast  # pylint: disable=import-outside-toplevel
                _vast_sdk = _vast.VastAI()
            except ImportError as e:
                raise ImportError(f'Fail to import dependencies for vast: {e}\n'
                                  'Try pip install "skypilot[vast]"') from None
        return func(*args, **kwargs)

    return wrapper


@import_package
def vast():
    """Return the vast package."""
    return _vast_sdk



"""Add job group columns to job_info table.

Adds:
- execution (TEXT) to job_info table: 'parallel' (job group) or 'serial'
  (pipeline/single job)

Note: cluster_name is not stored for job groups because it's deterministic
(computed from task name and job ID). Job groups don't support pools.

Revision ID: 013
Revises: 012
Create Date: 2025-12-29

"""
# pylint: disable=invalid-name
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa

from sky.dag import DagExecution
from sky.utils.db import db_utils

# revision identifiers, used by Alembic.
revision: str = '013'
down_revision: Union[str, Sequence[str], None] = '012'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade():
    """Add job group columns to job_info table."""
    with op.get_context().autocommit_block():
        # Add execution column to job_info table for execution mode:
        # 'parallel' (job group) or 'serial' (pipeline/single job)
        db_utils.add_column_to_table_alembic(
            'job_info',
            'execution',
            sa.Text(),
            server_default=DagExecution.SERIAL.value)
        # Update existing rows to have 'serial' execution mode
        op.execute(
            f'UPDATE job_info SET execution = \'{DagExecution.SERIAL.value}\' '
            'WHERE execution IS NULL')


def downgrade():
    """No downgrade logic."""
    pass



"""SDK functions for volumes."""
import json
import typing
from typing import List

from sky import exceptions
from sky import sky_logging
from sky.adaptors import common as adaptors_common
from sky.schemas.api import responses
from sky.server import common as server_common
from sky.server import versions
from sky.server.requests import payloads
from sky.usage import usage_lib
from sky.utils import annotations
from sky.utils import context
from sky.utils import ux_utils
from sky.volumes import volume as volume_lib

if typing.TYPE_CHECKING:
    import requests
else:
    requests = adaptors_common.LazyImport('requests')

logger = sky_logging.init_logger(__name__)


@context.contextual
@usage_lib.entrypoint
@server_common.check_server_healthy_or_start
@annotations.client_api
def apply(volume: volume_lib.Volume) -> server_common.RequestId[None]:
    """Creates or registers a volume.

    Example:
        .. code-block:: python

            import sky.volumes
            cfg = {
                'name': 'pvc',
                'type': 'k8s-pvc',
                'size': '100GB',
                'labels': {
                    'key': 'value',
                },
            }
            vol = sky.volumes.Volume.from_yaml_config(cfg)
            request_id = sky.volumes.apply(vol)
            sky.get(request_id)

            or

            import sky.volumes
            vol = sky.volumes.Volume(
                name='vol',
                type='runpod-network-volume',
                infra='runpod/ca/CA-MTL-1',
                size='100GB',
            )
            request_id = sky.volumes.apply(vol)
            sky.get(request_id)

    Args:
        volume: The volume to apply.

    Returns:
        The request ID of the apply request.
    """
    body = payloads.VolumeApplyBody(
        name=volume.name,
        volume_type=volume.type,
        cloud=volume.cloud,
        region=volume.region,
        zone=volume.zone,
        size=volume.size,
        config=volume.config,
        labels=volume.labels,
        use_existing=volume.use_existing,
    )
    response = server_common.make_authenticated_request(
        'POST', '/volumes/apply', json=json.loads(body.model_dump_json()))
    return server_common.get_request_id(response)


@context.contextual
@usage_lib.entrypoint
@server_common.check_server_healthy_or_start
@annotations.client_api
@versions.minimal_api_version(20)
def validate(volume: volume_lib.Volume) -> None:
    """Validates the volume.

    All validation is done on the server side.

    Args:
        volume: The volume to validate.

    Raises:
        ValueError: If the volume is invalid.
    """
    body = payloads.VolumeValidateBody(
        name=volume.name,
        volume_type=volume.type,
        infra=volume.infra,
        size=volume.size,
        config=volume.config,
        labels=volume.labels,
        use_existing=volume.use_existing,
    )
    response = server_common.make_authenticated_request(
        'POST', '/volumes/validate', json=json.loads(body.model_dump_json()))
    if response.status_code == 400:
        with ux_utils.print_exception_no_traceback():
            raise exceptions.deserialize_exception(
                response.json().get('detail'))


@context.contextual
@usage_lib.entrypoint
@server_common.check_server_healthy_or_start
@annotations.client_api
def ls(
    refresh: bool = False
) -> server_common.RequestId[List[responses.VolumeRecord]]:
    """Lists all volumes.

    Args:
        refresh: If True, refresh volume state from cloud APIs before returning.
            This makes the call slower but returns the most up-to-date data.
            If False (default), return cached data from the database which is
            updated periodically by the background daemon.

    Returns:
        The request ID of the list request.
    """
    params = {}
    if refresh:
        params['refresh'] = 'true'
    response = server_common.make_authenticated_request(
        'GET',
        '/volumes',
        params=params,
    )
    return server_common.get_request_id(response)


@context.contextual
@usage_lib.entrypoint
@server_common.check_server_healthy_or_start
@annotations.client_api
def delete(names: List[str],
           purge: bool = False) -> server_common.RequestId[None]:
    """Deletes volumes.

    Args:
        names: List of volume names to delete.
        purge: If True, delete the volume from the database even if the
          deletion API fails.

    Returns:
        The request ID of the delete request.
    """
    body = payloads.VolumeDeleteBody(names=names, purge=purge)
    response = server_common.make_authenticated_request(
        'POST', '/volumes/delete', json=json.loads(body.model_dump_json()))
    return server_common.get_request_id(response)



"""Utils shared between all of sky"""

import base64
import ctypes
import difflib
import enum
import functools
import gc
import getpass
import hashlib
import inspect
import json
import os
import platform
import random
import re
import socket
import subprocess
import sys
import time
import typing
from typing import Any, Callable, Dict, List, Optional, Union
import uuid

import jsonschema

from sky import exceptions
from sky import models
from sky import sky_logging
from sky.adaptors import common as adaptors_common
from sky.skylet import constants
from sky.usage import constants as usage_constants
from sky.utils import annotations
from sky.utils import context
from sky.utils import ux_utils
from sky.utils import validator

if typing.TYPE_CHECKING:
    import jinja2
    import psutil
else:
    jinja2 = adaptors_common.LazyImport('jinja2')
    psutil = adaptors_common.LazyImport('psutil')

USER_HASH_FILE = os.path.expanduser('~/.sky/user_hash')
USER_HASH_LENGTH = 8

# We are using base36 to reduce the length of the hash. 2 chars -> 36^2 = 1296
# possibilities. considering the final cluster name contains the prefix as well,
# we should be fine with 2 chars.
CLUSTER_NAME_HASH_LENGTH = 2

_COLOR_PATTERN = re.compile(r'\x1b[^m]*m')

_VALID_ENV_VAR_REGEX = '[a-zA-Z_][a-zA-Z0-9_]*'

logger = sky_logging.init_logger(__name__)


class ProcessStatus(enum.Enum):
    """Process status."""

    # The process is scheduled to run, but not started yet.
    SCHEDULED = 'SCHEDULED'

    # The process is running
    RUNNING = 'RUNNING'

    # The process is finished and succeeded
    SUCCEEDED = 'SUCCEEDED'

    # The process is interrupted
    INTERRUPTED = 'INTERRUPTED'

    # The process failed
    FAILED = 'FAILED'


@annotations.lru_cache(scope='request')
def get_usage_run_id() -> str:
    """Returns a unique run id for each 'run'.

    A run is defined as the lifetime of a process that has imported `sky`
    and has called its CLI or programmatic APIs. For example, two successive
    `sky launch` are two runs.
    """
    usage_run_id = os.getenv(usage_constants.USAGE_RUN_ID_ENV_VAR)
    if usage_run_id is not None:
        return usage_run_id
    return str(uuid.uuid4())


def is_valid_user_hash(user_hash: Optional[str]) -> bool:
    if user_hash is None:
        return False
    # Must start with a letter, followed by alphanumeric characters and hyphens
    # This covers both old hex format (e.g., "abc123") and new service account
    # format (e.g., "sa-abc123-token-xyz")
    return bool(re.match(r'^[a-zA-Z0-9][a-zA-Z0-9-]*$', user_hash))


def generate_user_hash() -> str:
    """Generates a unique user-machine specific hash."""
    hash_str = user_and_hostname_hash()
    user_hash = hashlib.md5(hash_str.encode()).hexdigest()[:USER_HASH_LENGTH]
    if not is_valid_user_hash(user_hash):
        # A fallback in case the hash is invalid.
        user_hash = uuid.uuid4().hex[:USER_HASH_LENGTH]
    return user_hash


def get_git_commit(path: Optional[str] = None) -> Optional[str]:
    try:
        result = subprocess.run(['git', 'rev-parse', 'HEAD'],
                                capture_output=True,
                                text=True,
                                cwd=path,
                                check=True)
        return result.stdout.strip()
    except subprocess.CalledProcessError:
        return None


def get_user_hash() -> str:
    """Returns a unique user-machine specific hash as a user id.

    We cache the user hash in a file to avoid potential user_name or
    hostname changes causing a new user hash to be generated.
    """
    user_hash = os.getenv(constants.USER_ID_ENV_VAR)
    if is_valid_user_hash(user_hash):
        assert user_hash is not None
        return user_hash

    if os.path.exists(USER_HASH_FILE):
        # Read from cached user hash file.
        with open(USER_HASH_FILE, 'r', encoding='utf-8') as f:
            # Remove invalid characters.
            user_hash = f.read().strip()
        if is_valid_user_hash(user_hash):
            return user_hash

    user_hash = generate_user_hash()
    set_user_hash_locally(user_hash)
    return user_hash


def set_user_hash_locally(user_hash: str) -> None:
    """Sets the user hash to local file."""
    os.makedirs(os.path.dirname(USER_HASH_FILE), exist_ok=True)
    with open(USER_HASH_FILE, 'w', encoding='utf-8') as f:
        f.write(user_hash)


def base36_encode(hex_str: str) -> str:
    """Converts a hex string to a base36 string."""
    int_value = int(hex_str, 16)

    def _base36_encode(num: int) -> str:
        if num == 0:
            return '0'
        alphabet = '0123456789abcdefghijklmnopqrstuvwxyz'
        base36 = ''
        while num != 0:
            num, i = divmod(num, 36)
            base36 = alphabet[i] + base36
        return base36

    return _base36_encode(int_value)


def check_cluster_name_is_valid(cluster_name: Optional[str]) -> None:
    """Errors out on invalid cluster names.

    Bans (including but not limited to) names that:
    - are digits-only
    - start with invalid character, like hyphen

    Raises:
        exceptions.InvalidClusterNameError: If the cluster name is invalid.
    """
    if cluster_name is None:
        return
    valid_regex = constants.CLUSTER_NAME_VALID_REGEX
    if re.fullmatch(valid_regex, cluster_name) is None:
        with ux_utils.print_exception_no_traceback():
            raise exceptions.InvalidClusterNameError(
                f'Cluster name "{cluster_name}" is invalid; '
                'ensure it is fully matched by regex (e.g., '
                'only contains letters, numbers and dash): '
                f'{valid_regex}')


def check_recipe_name_is_valid(recipe_name: Optional[str]) -> None:
    """Errors out on invalid recipe names.

    Recipe names must:
    - Start with a letter
    - Contain only letters, numbers, and dashes (no underscores or dots)
    - End with a letter or number
    - Be at most constants.RECIPE_NAME_MAX_LENGTH characters

    Raises:
        exceptions.InvalidRecipeNameError: If the recipe name is invalid.
    """
    if recipe_name is None:
        return
    if len(recipe_name) > constants.RECIPE_NAME_MAX_LENGTH:
        with ux_utils.print_exception_no_traceback():
            raise exceptions.InvalidRecipeNameError(
                f'Recipe name "{recipe_name}" is too long; '
                f'maximum length is {constants.RECIPE_NAME_MAX_LENGTH} '
                f'characters, got {len(recipe_name)}')
    valid_regex = constants.RECIPE_NAME_VALID_REGEX
    if re.fullmatch(valid_regex, recipe_name) is None:
        with ux_utils.print_exception_no_traceback():
            raise exceptions.InvalidRecipeNameError(
                f'Recipe name "{recipe_name}" is invalid; '
                'ensure it is fully matched by regex (e.g., '
                'only contains letters, numbers, and dashes).')


def make_cluster_name_on_cloud(display_name: str,
                               max_length: Optional[int] = 15,
                               add_user_hash: bool = True) -> str:
    """Generate valid cluster name on cloud that is unique to the user.

    This is to map the cluster name to a valid length and character set for
    cloud providers,
    - e.g. GCP limits the length of the cluster name to 35 characters. If the
    cluster name with user hash is longer than max_length:
      1. Truncate it to max_length - cluster_hash - user_hash_length.
      2. Append the hash of the cluster name
    - e.g. some cloud providers don't allow for uppercase letters, periods,
    or underscores, so we convert it to lower case and replace those
    characters with hyphens

    Args:
        display_name: The cluster name to be truncated, hashed, and
            transformed.
        max_length: The maximum length of the cluster name. If None, no
            truncation is performed.
        add_user_hash: Whether to append user hash to the cluster name.
    """

    cluster_name_on_cloud = re.sub(r'[._]', '-', display_name).lower()
    if display_name != cluster_name_on_cloud:
        logger.debug(
            f'The user specified cluster name {display_name} might be invalid '
            f'on the cloud, we convert it to {cluster_name_on_cloud}.')
    user_hash = ''
    if add_user_hash:
        user_hash = get_user_hash()
        user_hash = f'-{user_hash}'
    user_hash_length = len(user_hash)

    if (max_length is None or
            len(cluster_name_on_cloud) <= max_length - user_hash_length):
        return f'{cluster_name_on_cloud}{user_hash}'
    # -1 is for the dash between cluster name and cluster name hash.
    truncate_cluster_name_length = (max_length - CLUSTER_NAME_HASH_LENGTH - 1 -
                                    user_hash_length)
    truncate_cluster_name = cluster_name_on_cloud[:truncate_cluster_name_length]
    if truncate_cluster_name.endswith('-'):
        truncate_cluster_name = truncate_cluster_name.rstrip('-')
    assert truncate_cluster_name_length > 0, (cluster_name_on_cloud, max_length)
    display_name_hash = hashlib.md5(display_name.encode()).hexdigest()
    # Use base36 to reduce the length of the hash.
    display_name_hash = base36_encode(display_name_hash)
    return (f'{truncate_cluster_name}'
            f'-{display_name_hash[:CLUSTER_NAME_HASH_LENGTH]}{user_hash}')


def cluster_name_in_hint(cluster_name: str, cluster_name_on_cloud: str) -> str:
    if cluster_name_on_cloud.startswith(cluster_name):
        return repr(cluster_name)
    return f'{cluster_name!r} (name on cloud: {cluster_name_on_cloud!r})'


def get_global_job_id(job_timestamp: str,
                      cluster_name: Optional[str],
                      job_id: str,
                      task_id: Optional[int] = None,
                      is_managed_job: bool = False) -> str:
    """Returns a unique job run id for each job run.

    A job run is defined as the lifetime of a job that has been launched.
    """
    managed_job_str = 'managed-' if is_managed_job else ''
    _, sep, timestamp = job_timestamp.partition('sky-')
    job_timestamp = f'{sep}{managed_job_str}{timestamp}'
    global_job_id = f'{job_timestamp}_{cluster_name}_{job_id}'
    if task_id is not None:
        global_job_id += f'-{task_id}'
    return global_job_id


class Backoff:
    """Exponential backoff with jittering."""
    JITTER = 0.4

    def __init__(self,
                 initial_backoff: float = 5,
                 max_backoff_factor: int = 5,
                 multiplier: float = 1.6):
        self._initial = True
        self._backoff = 0.0
        self._initial_backoff = initial_backoff
        self._multiplier = multiplier
        self._max_backoff = max_backoff_factor * self._initial_backoff

    # https://github.com/grpc/grpc/blob/2d4f3c56001cd1e1f85734b2f7c5ce5f2797c38a/doc/connection-backoff.md
    # https://github.com/grpc/grpc/blob/5fc3ff82032d0ebc4bf252a170ebe66aacf9ed9d/src/core/lib/backoff/backoff.cc

    def current_backoff(self) -> float:
        """Backs off once and returns the current backoff in seconds."""
        if self._initial:
            self._initial = False
            self._backoff = min(self._initial_backoff, self._max_backoff)
        else:
            self._backoff = min(self._backoff * self._multiplier,
                                self._max_backoff)
        self._backoff += random.uniform(-self.JITTER * self._backoff,
                                        self.JITTER * self._backoff)
        return self._backoff


_CLIENT_COMMAND_KEY = 'client_command'
_CLIENT_ENTRYPOINT_KEY = 'client_entrypoint'
_USING_REMOTE_API_SERVER_KEY = 'using_remote_api_server'
_USER_KEY = 'user'
_REQUEST_ID_KEY = 'request_id'


def set_request_context(client_entrypoint: Optional[str],
                        client_command: Optional[str],
                        using_remote_api_server: bool,
                        user: Optional['models.User'], request_id: str) -> None:
    """Override the current client entrypoint and command.

    This is useful when we are on the SkyPilot API server side and we have a
    client entrypoint and command from the client.
    """
    # This function will be called in process executor and coroutine executor.
    # context.set_context_var ensures the context is safe in both cases.
    context.set_context_var(_CLIENT_ENTRYPOINT_KEY, client_entrypoint)
    context.set_context_var(_CLIENT_COMMAND_KEY, client_command)
    context.set_context_var(_USING_REMOTE_API_SERVER_KEY,
                            using_remote_api_server)
    context.set_context_var(_USER_KEY, user)
    context.set_context_var(_REQUEST_ID_KEY, request_id)


def get_current_request_id() -> str:
    """Returns the current request id."""
    value = context.get_context_var('request_id')
    if value is not None:
        return value
    return 'dummy-request-id'


def get_current_command() -> str:
    """Returns the command related to this operation.

    Normally uses get_pretty_entry_point(), but will use the client command on
    the server side.
    """
    value = context.get_context_var(_CLIENT_COMMAND_KEY)
    if value is not None:
        return value
    return get_pretty_entrypoint_cmd()


def get_current_user() -> 'models.User':
    """Returns the user in current server session."""
    value = context.get_context_var(_USER_KEY)
    if value is not None:
        return value
    return models.User.get_current_user()


def get_current_user_name() -> str:
    """Returns the user name in current server session."""
    name = get_current_user().name
    assert name is not None
    return name


def get_local_user_name() -> str:
    """Returns the user name in local environment.

    This is for backward compatibility where anonymous access is implicitly
    allowed when no authentication method at server-side is configured and
    the username from client environment variable will be used to identify the
    user.
    """
    name = os.getenv(constants.USER_ENV_VAR, getpass.getuser())
    assert name is not None
    return name


def set_current_user(user: 'models.User'):
    """Sets the current user."""
    context.set_context_var('user', user)


def get_current_client_entrypoint(server_entrypoint: str) -> str:
    """Returns the current client entrypoint.

    Gets the client entrypoint from the context, if it is not set, returns the
    server entrypoint.
    """
    value = context.get_context_var(_CLIENT_ENTRYPOINT_KEY)
    if value is not None:
        return value
    return server_entrypoint


def get_using_remote_api_server() -> bool:
    """Returns whether the API server is remote."""
    if os.getenv(constants.USING_REMOTE_API_SERVER_ENV_VAR) is not None:
        return os.getenv(constants.USING_REMOTE_API_SERVER_ENV_VAR,
                         '').lower() in ('true', '1')
    value = context.get_context_var(_USING_REMOTE_API_SERVER_KEY)
    if value is not None:
        return value
    # This gets the right status for the local client.
    # TODO(zhwu): This is to prevent circular import. We should refactor this.
    # pylint: disable=import-outside-toplevel
    from sky.server import common as server_common
    return not server_common.is_api_server_local()


def get_pretty_entrypoint_cmd() -> str:
    """Returns the prettified entry point of this process (sys.argv).

    Example return values:
        $ sky launch app.yaml  # 'sky launch app.yaml'
        $ python examples/app.py  # 'app.py'
    """
    argv = sys.argv
    basename = os.path.basename(argv[0])
    if basename == 'sky':
        # Turn '/.../anaconda/envs/py36/bin/sky' into 'sky', but keep other
        # things like 'examples/app.py'.
        argv[0] = basename

    # Redact sensitive values from secrets arguments
    argv = _redact_secrets_values(argv)

    return ' '.join(argv)


def read_last_n_lines(file_path: str,
                      n: int,
                      chunk_size: int = 8192,
                      encoding: str = 'utf-8',
                      errors: str = 'replace') -> List[str]:
    """Read the last N lines of a file.

    Args:
        file_path: Path to the file to read.
        n: Number of lines to read from the end of the file.
        chunk_size: Size of chunks in bytes.
        encoding: Encoding to use when decoding binary chunks.
        errors: Error handling for decode errors (e.g., 'replace', 'ignore').

    Returns:
        A list of the last N lines, preserving newlines where applicable.
    """

    assert n >= 0, f'n must be non-negative. Got {n}'
    assert chunk_size > 0, f'chunk_size must be positive. Got {chunk_size}'
    assert os.path.exists(file_path), f'File not found: {file_path}'

    if n == 0:
        return []

    try:
        with open(file_path, 'rb') as f:
            # Start reading from the end of the file
            f.seek(0, os.SEEK_END)
            file_size = f.tell()
            if file_size == 0:
                return []

            pos = file_size
            lines_found = 0
            chunks = []

            # Read backwards in chunks until we've found at least n newlines
            while pos > 0 and lines_found <= n:
                read_size = min(chunk_size, pos)
                pos -= read_size
                f.seek(pos)
                chunk = f.read(read_size)
                chunks.append(chunk)
                lines_found += chunk.count(b'\n')

            # Combine all chunks in reverse order since we read backwards
            full_bytes = b''.join(reversed(chunks))

            # Split by newline byte. Note: this handles '\n' endings.
            all_lines = full_bytes.split(b'\n')

            # Handle edge case: if file ends with a newline, last element is b''
            if all_lines and all_lines[-1] == b'':
                result_bytes = all_lines[-n - 1:-1]
            else:
                result_bytes = all_lines[-n:]

            # Decode each line and normalize CR/LF endings
            decoded_lines = [
                line.decode(encoding, errors=errors).rstrip('\r') + '\n'
                for line in result_bytes[:-1]
            ]

            # Decode the final line — only add newline if it was present
            last_line = result_bytes[-1].decode(encoding,
                                                errors=errors).rstrip('\r')
            decoded_lines.append(last_line)

            return decoded_lines

    except OSError as e:
        with ux_utils.print_exception_no_traceback():
            raise RuntimeError(
                f'Failed to read last {n} lines from {file_path}: {e}') from e


def _redact_secrets_values(argv: List[str]) -> List[str]:
    """Redact sensitive values from --secret arguments.

    Args:
        argv: Command line arguments

    Returns:
        Modified argv with redacted --secret values, or original argv if any
        error

    Examples:
        ['sky', 'launch', '--secret', 'HF_TOKEN=secret'] ->
        ['sky', 'launch', '--secret', 'HF_TOKEN=<redacted>']

        ['sky', 'launch', '--secret=HF_TOKEN=secret'] ->
        ['sky', 'launch', '--secret=HF_TOKEN=<redacted>']

        ['sky', 'launch', '--secret', 'HF_TOKEN'] ->
        ['sky', 'launch', '--secret', 'HF_TOKEN'] (no change)
    """
    try:
        if not argv:
            return argv or []

        result = []
        i = 0

        while i < len(argv):
            arg = argv[i]

            # Ensure arg is a string
            if not isinstance(arg, str):
                result.append(arg)
                i += 1
                continue

            if arg == '--secret' and i + 1 < len(argv):
                result.append(arg)
                next_arg = argv[i + 1]
                # Ensure next_arg is a string and handle redaction safely
                if isinstance(next_arg, str):
                    redacted = re.sub(r'^([^=]+)=.*', r'\1=<redacted>',
                                      next_arg)
                    result.append(redacted)
                else:
                    result.append(next_arg)
                i += 2
            elif arg.startswith('--secret='):
                # Redact only if there's a value after the key
                redacted = re.sub(r'^(--secret=[^=]+)=.*', r'\1=<redacted>',
                                  arg)
                result.append(redacted)
                i += 1
            else:
                result.append(arg)
                i += 1

        return result
    except Exception:  # pylint: disable=broad-except
        # If anything goes wrong with redaction, return original argv
        # This ensures the command can still execute
        return argv or []


def user_and_hostname_hash() -> str:
    """Returns a string containing <user>-<hostname hash last 4 chars>.

    For uniquefying user clusters on shared-account cloud providers. Also used
    for AWS security group.

    Using uuid.getnode() instead of gethostname() is incorrect; observed to
    collide on Macs.

    NOTE: BACKWARD INCOMPATIBILITY NOTES

    Changing this string will render AWS clusters shown in `sky status`
    unreusable and potentially cause leakage:

    - If a cluster is STOPPED, any command restarting it (`sky launch`, `sky
      start`) will launch a NEW cluster.
    - If a cluster is UP, a `sky launch` command reusing it will launch a NEW
      cluster. The original cluster will be stopped and thus leaked from Sky's
      perspective.
    - `sky down/stop/exec` on these pre-change clusters still works, if no new
      clusters with the same name have been launched.

    The reason is AWS security group names are derived from this string, and
    thus changing the SG name makes these clusters unrecognizable.
    """
    hostname_hash = hashlib.md5(socket.gethostname().encode()).hexdigest()[-4:]
    return f'{getpass.getuser()}-{hostname_hash}'


def make_decorator(cls, name_or_fn: Union[str, Callable],
                   **ctx_kwargs) -> Callable:
    """Make the cls a decorator.

    class cls:
        def __init__(self, name, **kwargs):
            pass
        def __enter__(self):
            pass
        def __exit__(self, exc_type, exc_value, traceback):
            pass

    Args:
        name_or_fn: The name of the event or the function to be wrapped.
        message: The message attached to the event.
    """
    if isinstance(name_or_fn, str):

        def _wrapper(f):

            @functools.wraps(f)
            def _record(*args, **kwargs):
                with cls(name_or_fn, **ctx_kwargs):
                    return f(*args, **kwargs)

            return _record

        return _wrapper
    else:
        if not inspect.isfunction(name_or_fn):
            raise ValueError(
                'Should directly apply the decorator to a function.')

        @functools.wraps(name_or_fn)
        def _record(*args, **kwargs):
            f = name_or_fn
            func_name = getattr(f, '__qualname__', f.__name__)
            module_name = getattr(f, '__module__', '')
            if module_name:
                full_name = f'{module_name}.{func_name}'
            else:
                full_name = func_name
            with cls(full_name, **ctx_kwargs):
                return f(*args, **kwargs)

        return _record


def retry(method, max_retries=3, initial_backoff=1):
    """Retry a function up to max_retries times with backoff between retries."""

    @functools.wraps(method)
    def method_with_retries(*args, **kwargs):
        backoff = Backoff(initial_backoff)
        try_count = 0
        while try_count < max_retries:
            try:
                return method(*args, **kwargs)
            except Exception as e:  # pylint: disable=broad-except
                try_count += 1
                if try_count < max_retries:
                    logger.warning(f'Caught {e}. Retrying.')
                    time.sleep(backoff.current_backoff())
                else:
                    raise

    return method_with_retries


def class_fullname(cls, skip_builtins: bool = True):
    """Get the full name of a class.

    Example:
        >>> e = sky.exceptions.FetchClusterInfoError()
        >>> class_fullname(e.__class__)
        'sky.exceptions.FetchClusterInfoError'

    Args:
        cls: The class to get the full name.

    Returns:
        The full name of the class.
    """
    module_name = getattr(cls, '__module__', '')
    if not module_name or (module_name == 'builtins' and skip_builtins):
        return cls.__name__
    return f'{cls.__module__}.{cls.__name__}'


def format_exception(e: Union[Exception, SystemExit, KeyboardInterrupt],
                     use_bracket: bool = False) -> str:
    """Format an exception to a string.

    Args:
        e: The exception to format.

    Returns:
        A string that represents the exception.
    """
    if use_bracket:
        return f'[{class_fullname(e.__class__)}] {e}'
    return f'{class_fullname(e.__class__)}: {e}'


def remove_color(s: str):
    """Remove color from a string.

    Args:
        s: The string to remove color.

    Returns:
        A string without color.
    """
    return _COLOR_PATTERN.sub('', s)


def remove_file_if_exists(path: Optional[str]):
    """Delete a file if it exists.

    Args:
        path: The path to the file.
    """
    if path is None:
        return
    try:
        os.remove(path)
    except FileNotFoundError:
        logger.debug(f'Tried to remove {path} but failed to find it. Skip.')
        pass


def is_wsl() -> bool:
    """Detect if running under Windows Subsystem for Linux (WSL)."""
    return 'microsoft' in platform.uname().release.lower()


def find_free_port(start_port: int) -> int:
    """Finds first free local port starting with 'start_port'.

    Returns: a free local port.

    Raises:
      OSError: If no free ports are available.
    """
    for port in range(start_port, 65535):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.bind(('', port))
                return port
            except OSError as e:
                logger.debug(f'Error binding port {port}: {e}')
                pass
    raise OSError('No free ports available.')


def is_valid_env_var(name: str) -> bool:
    """Checks if the task environment variable name is valid."""
    return bool(re.fullmatch(_VALID_ENV_VAR_REGEX, name))


def format_float(num: Union[float, int], precision: int = 1) -> str:
    """Formats a float to not show decimal point if it is a whole number

    If it is not a whole number, it will show upto precision decimal point."""
    if isinstance(num, int):
        return str(num)
    return '{:.0f}'.format(num) if num.is_integer() else f'{num:.{precision}f}'


def validate_schema(obj, schema, err_msg_prefix='', skip_none=True):
    """Validates an object against a given JSON schema.

    Args:
        obj: The object to validate.
        schema: The JSON schema against which to validate the object.
        err_msg_prefix: The string to prepend to the error message if
          validation fails.
        skip_none: If True, removes fields with value None from the object
          before validation. This is useful for objects that will never contain
          None because yaml.safe_load() loads empty fields as None.

    Raises:
        ValueError: if the object does not match the schema.
    """
    if skip_none:
        obj = {k: v for k, v in obj.items() if v is not None}
    err_msg = None
    try:
        validator.get_schema_validator()(schema).validate(obj)
    except jsonschema.ValidationError as e:
        if e.validator == 'additionalProperties':
            if tuple(e.schema_path) == ('properties', 'envs',
                                        'additionalProperties'):
                # Hack. Here the error is Task.envs having some invalid keys. So
                # we should not print "unsupported field".
                #
                # This will print something like:
                # 'hello world' does not match any of the regexes: <regex>
                err_msg = (err_msg_prefix +
                           'The `envs` field contains invalid keys:\n' +
                           e.message)
            else:
                err_msg = err_msg_prefix
                assert isinstance(e.schema, dict), 'Schema must be a dictionary'
                known_fields = set(e.schema.get('properties', {}).keys())
                assert isinstance(e.instance,
                                  dict), 'Instance must be a dictionary'
                for field in e.instance:
                    if field not in known_fields:
                        most_similar_field = difflib.get_close_matches(
                            field, known_fields, 1)
                        if most_similar_field:
                            err_msg += (f'Instead of {field!r}, did you mean '
                                        f'{most_similar_field[0]!r}?')
                        else:
                            err_msg += f'Found unsupported field {field!r}.'
        else:
            message = e.message
            # Object in jsonschema is represented as dict in Python. Replace
            # 'object' with 'dict' for better readability.
            message = message.replace('type \'object\'', 'type \'dict\'')
            # Example e.json_path value: '$.resources'
            err_msg = (err_msg_prefix + message +
                       f'. Check problematic field(s): {e.json_path}')

    if err_msg:
        with ux_utils.print_exception_no_traceback():
            raise exceptions.InvalidSkyPilotConfigError(err_msg)


def get_cleaned_username(username: str = '') -> str:
    """Cleans the username. Underscores are allowed, as we will
     handle it when mapping to the cluster_name_on_cloud in
     common_utils.make_cluster_name_on_cloud.

    Clean up includes:
     1. Making all characters lowercase
     2. Removing any non-alphanumeric characters (excluding hyphens and
        underscores)
     3. Removing any numbers and/or hyphens at the start of the username.
     4. Removing any hyphens at the end of the username
     5. Truncate the username to 63 characters, as requested by GCP labels

    Dots are removed due to: https://cloud.google.com/compute/docs/labeling-resources#requirements # pylint: disable=line-too-long

    e.g. 1SkY-PiLot2- becomes sky-pilot2

    Returns:
      A cleaned username.
    """
    username = username or get_current_user_name()
    username = username.lower()
    username = re.sub(r'[^a-z0-9-_]', '', username)
    username = re.sub(r'^[0-9-]+', '', username)
    username = re.sub(r'-$', '', username)
    username = username[:63]
    return username


def fill_template(template_name: str, variables: Dict[str, Any],
                  output_path: str) -> None:
    """Create a file from a Jinja template and return the filename."""
    assert template_name.endswith('.j2'), template_name
    root_dir = os.path.dirname(os.path.dirname(__file__))
    template_path = os.path.join(root_dir, 'templates', template_name)
    if not os.path.exists(template_path):
        raise FileNotFoundError(f'Template "{template_name}" does not exist.')
    with open(template_path, 'r', encoding='utf-8') as fin:
        template = fin.read()
    output_path = os.path.abspath(os.path.expanduser(output_path))
    os.makedirs(os.path.dirname(output_path), exist_ok=True)

    # Write out yaml config.
    j2_template = jinja2.Template(template)
    content = j2_template.render(**variables)
    with open(output_path, 'w', encoding='utf-8') as fout:
        fout.write(content)


def deprecated_function(
        func: Callable,
        name: str,
        deprecated_name: str,
        removing_version: str,
        override_argument: Optional[Dict[str, Any]] = None) -> Callable:
    """Decorator for creating deprecated functions, for backward compatibility.

    It will result in a warning being emitted when the function is used.
    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        override_argument_str = ''
        if override_argument:
            override_argument_str = ', '.join(
                f'{k}={v}' for k, v in override_argument.items())
        logger.warning(
            f'Call to deprecated function {deprecated_name}, which will be '
            f'removed in {removing_version}. Please use '
            f'{name}({override_argument_str}) instead.')
        return func(*args, **kwargs)

    return new_func


def truncate_long_string(s: str,
                         max_length: int = 35,
                         truncate_middle: bool = False) -> str:
    """Truncate a string to a maximum length.

    Args:
        s: String to truncate.
        max_length: Maximum length of the truncated string.
        truncate_middle: Whether to truncate in the middle of the string.
            If True, the middle part of the string is replaced with '...'.
            If False, truncation happens at the end preserving whole words.

    Returns:
        Truncated string.
    """
    if len(s) <= max_length:
        return s

    if truncate_middle:
        # Reserve 3 characters for '...'
        if max_length <= 3:
            return '...'

        # Calculate how many characters to keep from beginning and end
        half_length = (max_length - 3) // 2
        remainder = (max_length - 3) % 2

        # Keep one more character at the beginning if max_length - 3 is odd
        start_length = half_length + remainder
        end_length = half_length

        # When end_length is 0, just show the start part and '...'
        if end_length == 0:
            return s[:start_length] + '...'
        return s[:start_length] + '...' + s[-end_length:]

    # Original end-truncation logic
    splits = s.split(' ')
    if len(splits[0]) > max_length:
        return splits[0][:max_length] + '...'  # Use '…'?
    # Truncate on word boundary.
    i = 0
    total = 0
    for i, part in enumerate(splits):
        total += len(part)
        if total >= max_length:
            break
    prefix = ' '.join(splits[:i])
    if len(prefix) < max_length:
        prefix += s[len(prefix):max_length]
    return prefix + '...'


def hash_file(path: str, hash_alg: str) -> 'hashlib._Hash':
    # In python 3.11, hashlib.file_digest is available, but for <3.11 we have to
    # do it manually.
    # This implementation is simplified from the implementation in CPython.
    # TODO(cooperc): Use hashlib.file_digest once we move to 3.11+.
    # Beware of f.read() as some files may be larger than memory.
    with open(path, 'rb') as f:
        file_hash = hashlib.new(hash_alg)
        buf = bytearray(2**18)
        view = memoryview(buf)
        while True:
            size = f.readinto(buf)
            if size == 0:
                # EOF
                break
            file_hash.update(view[:size])
        return file_hash


def is_port_available(port: int, reuse_addr: bool = True) -> bool:
    """Check if a TCP port is available for binding on localhost.

    Args:
        port: The port number to check.
        reuse_addr: If True, sets SO_REUSEADDR socket option to allow reusing
            ports in TIME_WAIT state. Servers like multiprocessing.Manager set
            SO_REUSEADDR by default to accelerate restart. The option should be
            coordinated in check.

    Returns:
        bool: True if the port is available for binding, False otherwise.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        if reuse_addr:
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            s.bind(('localhost', port))
            return True
        except OSError:
            return False


def get_cpu_count() -> int:
    """Get the number of CPUs, with cgroup awareness."""
    # This env-var is kept since it is still useful for limiting the resource
    # of SkyPilot in non-containerized environments.
    cpu_count = os.getenv('SKYPILOT_POD_CPU_CORE_LIMIT')
    if cpu_count is not None:
        try:
            return int(float(cpu_count))
        except ValueError as e:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'Failed to parse the number of CPUs from {cpu_count}'
                ) from e
    return _cpu_count()


def get_mem_size_gb() -> float:
    """Get the memory size in GB, with cgroup awareness."""
    mem_size = os.getenv('SKYPILOT_POD_MEMORY_GB_LIMIT')
    if mem_size is not None:
        try:
            return float(mem_size)
        except ValueError as e:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'Failed to parse the memory size from {mem_size} (GB)'
                ) from e
    mem_size = os.getenv('SKYPILOT_POD_MEMORY_BYTES_LIMIT')
    if mem_size is not None:
        try:
            return float(mem_size) / (1024**3)
        except ValueError as e:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'Failed to parse the memory size from {mem_size} (bytes)'
                ) from e
    return _mem_size_gb()


def _cpu_count() -> int:
    # host cpu cores (logical)
    cpu = psutil.cpu_count()
    # cpu affinity on Linux
    if hasattr(os, 'sched_getaffinity'):
        # just for safe, length of CPU set should always <= logical cpu cores
        cpu = min(cpu, len(os.sched_getaffinity(0)))
    cgroup_cpu = _get_cgroup_cpu_limit()
    if cgroup_cpu is not None:
        cpu = min(cpu, int(cgroup_cpu))
    return cpu


def _mem_size_gb() -> float:
    # host memory limit
    mem = psutil.virtual_memory().total
    cgroup_mem = _get_cgroup_memory_limit()
    if cgroup_mem is not None:
        mem = min(mem, cgroup_mem)
    return mem / (1024**3)


# Refer to:
# - https://docs.kernel.org/admin-guide/cgroup-v1/index.html
# - https://docs.kernel.org/admin-guide/cgroup-v2.html
# for the standards of handler files in cgroupv1 and v2.
# Since all those paths are well-known standards that are unlikely to change,
# we use string literals instead of defining extra constants.
def _get_cgroup_cpu_limit() -> Optional[float]:
    """Return cpu limit from cgroups in cores.

    Returns:
        The cpu limit in cores as a float (can be fractional), or None if there
        is no limit in cgroups.
    """
    try:
        if _is_cgroup_v2():
            with open('/sys/fs/cgroup/cpu.max', 'r', encoding='utf-8') as f:
                quota_str, period_str = f.read().strip().split()
                if quota_str == 'max':
                    return None
                quota = float(quota_str)
                period = float(period_str)
                return quota / period if quota > 0 else None
        else:
            # cgroup v1
            with open('/sys/fs/cgroup/cpu/cpu.cfs_quota_us',
                      'r',
                      encoding='utf-8') as f:
                quota = float(f.read().strip())
            with open('/sys/fs/cgroup/cpu/cpu.cfs_period_us',
                      'r',
                      encoding='utf-8') as f:
                period = float(f.read().strip())
            # Return unlimited if cpu quota is not set.
            # Note that we do not use cpu.shares since it is a relative weight
            # instead of a hard limit. It is okay to get CPU throttling under
            # high contention. And unlimited enables the server to use as much
            # CPU as available if there is no contention.
            return quota / period if (quota > 0 and period > 0) else None
    except (OSError, ValueError):
        return None


def _get_cgroup_memory_limit() -> Optional[int]:
    """Return memory limit from cgroups in bytes.

    Returns:
        The memory limit in bytes, or None if there is no limit in cgroups.
    """
    try:
        path = ('/sys/fs/cgroup/memory.max' if _is_cgroup_v2() else
                '/sys/fs/cgroup/memory/memory.limit_in_bytes')
        with open(path, 'r', encoding='utf-8') as f:
            value = f.read().strip()
            if value == 'max' or not value:
                return None
            limit = int(value)
            return limit if limit > 0 else None
    except (OSError, ValueError):
        return None


def _is_cgroup_v2() -> bool:
    """Return True if the environment is running cgroup v2."""
    return os.path.isfile('/sys/fs/cgroup/cgroup.controllers')


def removeprefix(string: str, prefix: str) -> str:
    if string.startswith(prefix):
        return string[len(prefix):]
    return string


def release_memory():
    """Release the process memory"""
    # Do the best effort to release the python heap and let malloc_trim
    # be more efficient.
    try:
        gc.collect()
        if sys.platform.startswith('linux'):
            # Will fail on musl (alpine), but at least it works on our
            # official docker images.
            libc = ctypes.CDLL('libc.so.6')
            return libc.malloc_trim(0)
        return 0
    except Exception as e:  # pylint: disable=broad-except
        logger.error(f'Failed to release memory: '
                     f'{format_exception(e)}')
        return 0


def base64_url_encode(data: bytes) -> str:
    """Base64url encode data without padding.

    Uses URL-safe alphabet (- and _ instead of + and /) and strips padding
    to avoid URL encoding issues with = characters.
    """
    return base64.urlsafe_b64encode(data).rstrip(b'=').decode('ascii')


def compute_code_challenge(code_verifier: str) -> str:
    """Compute a code_challenge from code_verifier using SHA256.

    Used in the CLI login flow for CSRF protection. The CLI generates a
    random code_verifier, computes the challenge, and sends the challenge
    to the server. Later, the CLI proves it initiated the request by
    providing the original verifier which the server hashes to verify.
    """
    digest = hashlib.sha256(code_verifier.encode('utf-8')).digest()
    return base64_url_encode(digest)


def merge_node_names_lineage(
        existing_json: Optional[str],
        current_names: Optional[List[str]]) -> Optional[str]:
    """Merge current node names into the existing node name lineage.

    The node_names column stores a JSON list of lists. Each inner list
    represents one node (head first, then workers) and contains the
    lineage of names that have been assigned to that node over time
    (e.g. across recovery events). Each inner list is capped at
    constants.MAX_NODE_NAME_LINEAGE entries.

    Args:
        existing_json: JSON string of list-of-lists from the DB, or None.
        current_names: List of current node names (head first), or None.

    Returns:
        JSON string of list-of-lists with lineage updated, or None.
    """
    if current_names is None:
        return existing_json

    existing: List[List[str]] = []
    if existing_json is not None:
        try:
            parsed = json.loads(existing_json)
            if isinstance(parsed, list):
                existing = parsed
        except (json.JSONDecodeError, TypeError):
            pass

    result: List[List[str]] = []
    for i, name in enumerate(current_names):
        if i < len(existing) and isinstance(existing[i], list):
            lineage = list(existing[i])
            # Only append if name is valid and different from last entry
            if name and (not lineage or lineage[-1] != name):
                lineage.append(name)
            # Cap at max lineage length
            if len(lineage) > constants.MAX_NODE_NAME_LINEAGE:
                lineage = lineage[-constants.MAX_NODE_NAME_LINEAGE:]
        else:
            lineage = [name] if name else []
        result.append(lineage)

    return json.dumps(result) if result else None


def get_display_node_names(node_names_json: Optional[str]) -> Optional[str]:
    """Extract display node names from the lineage JSON.

    Takes the last entry from each inner list (the most recent name for
    each node) and returns them as a comma-separated string.

    Args:
        node_names_json: JSON string of list-of-lists, or None.

    Returns:
        Comma-separated string of current node names, or None.
    """
    if node_names_json is None:
        return None

    try:
        lineage = json.loads(node_names_json)
        if not isinstance(lineage, list):
            # Backward compat: might be old comma-separated format
            return node_names_json
        names = []
        for node_lineage in lineage:
            if isinstance(node_lineage, list) and node_lineage:
                names.append(node_lineage[-1])
            elif isinstance(node_lineage, str):
                # Backward compat: flat list of strings
                names.append(node_lineage)
        return ','.join(names) if names else None
    except (json.JSONDecodeError, TypeError):
        # Backward compat: return as-is if not valid JSON
        return node_names_json



"""Utilities for building middlewares."""
import enum
import http
from typing import Type

import fastapi
import starlette.middleware.base
import starlette.types

from sky import sky_logging

logger = sky_logging.init_logger(__name__)


class WebSocketDecision(enum.Enum):
    ACCEPT = 'accept'
    UNAUTHORIZED = 'unauthorized'
    FORBIDDEN = 'forbidden'
    ERROR = 'error'


def websocket_aware(
        middleware_cls: Type[starlette.middleware.base.BaseHTTPMiddleware]):
    """Decorator to adapt BaseHTTPMiddleware to handle WebSockets.

    It assembles an HTTP-style request like the HTTP upgrade request during
    websocket handshake and then delegates it to the real HTTP middleware.
    The websocket connection will be rejected if the HTTP middleware returns
    a 4xx or 5xx status code.

    Note: for websocket connection, the mutation made by the underlying HTTP
    middleware on the request and response will be discarded.
    """

    class WebSocketAwareMiddleware:
        """WebSocket-aware middleware wrapper."""

        def __init__(self, app: starlette.types.ASGIApp, *args, **kwargs):
            self.app = app
            self.middleware = middleware_cls(app, *args, **kwargs)

        async def __call__(self, scope: starlette.types.Scope,
                           receive: starlette.types.Receive,
                           send: starlette.types.Send):
            scope_type = scope.get('type')
            if scope_type == 'websocket':
                await self._handle_websocket(scope, receive, send)
            else:
                # Delegate other scopes to the underlying HTTP middleware.
                await self.middleware(scope, receive, send)

        async def dispatch(
                self, request: fastapi.Request,
                call_next: starlette.middleware.base.RequestResponseEndpoint):
            """Implement dispatch method to keep compatibility."""
            return await self.middleware.dispatch(request, call_next)

        async def _handle_websocket(self, scope: starlette.types.Scope,
                                    receive: starlette.types.Receive,
                                    send: starlette.types.Send):
            """Handle websocket connection by delegating to HTTP middleware."""
            decision = await self._run_websocket_dispatch(scope)
            if decision == WebSocketDecision.ACCEPT:
                await self.app(scope, receive, send)
            elif decision == WebSocketDecision.UNAUTHORIZED:
                await send({
                    'type': 'websocket.close',
                    'code': 4401,
                    'reason': 'Unauthorized',
                })
            elif decision == WebSocketDecision.FORBIDDEN:
                await send({
                    'type': 'websocket.close',
                    'code': 4403,
                    'reason': 'Forbidden',
                })
            else:
                await send({
                    'type': 'websocket.close',
                    'code': 1011,
                    'reason': 'Internal Server Error',
                })

        async def _run_websocket_dispatch(
                self, scope: starlette.types.Scope) -> WebSocketDecision:
            http_scope = self._build_http_scope(scope)
            http_receive = self._http_receive_adapter()
            request = fastapi.Request(http_scope, receive=http_receive)
            call_next_called = False
            stub_response = fastapi.Response(status_code=http.HTTPStatus.OK)

            async def call_next(req):
                del req
                # Capture whether call_next() is called in the underlying
                # HTTP middleware to determine if we can proceed with current
                # websocket connection.
                nonlocal call_next_called
                call_next_called = True
                return stub_response

            try:
                response = await self.dispatch(request, call_next)
            except Exception as e:  # pylint: disable=broad-except
                logger.error('Exception occurred in middleware dispatch for '
                             f'WebSocket scope: {e}')
                return WebSocketDecision.ERROR

            if response is None:
                response = stub_response

            status_code = response.status_code

            if call_next_called and 200 <= status_code < 400:
                return WebSocketDecision.ACCEPT
            if status_code == http.HTTPStatus.UNAUTHORIZED:
                return WebSocketDecision.UNAUTHORIZED
            if status_code == http.HTTPStatus.FORBIDDEN:
                return WebSocketDecision.FORBIDDEN
            return WebSocketDecision.ERROR

        @staticmethod
        def _build_http_scope(
                scope: starlette.types.Scope) -> starlette.types.Scope:
            state = scope.setdefault('state', {})
            scheme = scope.get('scheme', 'ws')
            if scheme == 'ws':
                http_scheme = 'http'
            elif scheme == 'wss':
                http_scheme = 'https'
            else:
                http_scheme = scheme
            http_scope = dict(scope)
            http_scope['type'] = 'http'
            http_scope['scheme'] = http_scheme
            http_scope['method'] = 'GET'
            http_scope['http_version'] = scope.get('http_version', '1.1')
            http_scope['state'] = state
            return http_scope

        @staticmethod
        def _http_receive_adapter() -> starlette.types.Receive:
            """Adapter thatmimics the sequence produced by Starlette for an HTTP
            request: a single http.request event followed by a http.disconnect
            """
            sent = False

            async def receive():
                nonlocal sent
                if not sent:
                    sent = True
                    return {
                        'type': 'http.request',
                        'body': b'',
                        'more_body': False,
                    }
                return {
                    'type': 'http.disconnect',
                }

            return receive

    WebSocketAwareMiddleware.__name__ = middleware_cls.__name__
    WebSocketAwareMiddleware.__qualname__ = middleware_cls.__qualname__
    WebSocketAwareMiddleware.__module__ = middleware_cls.__module__
    WebSocketAwareMiddleware.__doc__ = middleware_cls.__doc__
    return WebSocketAwareMiddleware



"""This module helps generating timelines of an application.

The timeline follows the trace event format defined here:
https://docs.google.com/document/d/1CvAClvFfyA5R-PhYUmn5OOQtYMH4h6I0nSsKchNAySU/preview
"""  # pylint: disable=line-too-long
import atexit
import json
import os
import threading
import time
import traceback
from typing import Callable, Optional, Union

from sky.utils import common_utils

_events = []


def _get_events_file_path():
    return os.environ.get('SKYPILOT_TIMELINE_FILE_PATH')


class Event:
    """Record an event.

    Args:
        name: The name of the event.
        message: The message attached to the event.
    """

    def __init__(self, name: str, message: Optional[str] = None):
        self._skipped = False
        if not _get_events_file_path():
            self._skipped = True
            return
        self._name = name
        self._message = message
        # See the module doc for the event format.
        self._event = {
            'name': self._name,
            'cat': 'event',
            'pid': str(os.getpid()),
            'tid': str(threading.current_thread().ident),
            'args': {
                'message': self._message
            }
        }
        if self._message is not None:
            self._event['args'] = {'message': self._message}

    def begin(self):
        if self._skipped:
            return
        event_begin = self._event.copy()
        event_begin.update({
            'ph': 'B',
            'ts': f'{time.time() * 10 ** 6: .3f}',
        })
        event_begin['args'] = {'stack': '\n'.join(traceback.format_stack())}
        if self._message is not None:
            event_begin['args'][
                'message'] = self._message  # type: ignore[index]
        _events.append(event_begin)

    def end(self):
        if self._skipped:
            return
        event_end = self._event.copy()
        event_end.update({
            'ph': 'E',
            'ts': f'{time.time() * 10 ** 6: .3f}',
        })
        if self._message is not None:
            event_end['args'] = {'message': self._message}
        _events.append(event_end)

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end()


def event(name_or_fn: Union[str, Callable], message: Optional[str] = None):
    return common_utils.make_decorator(Event, name_or_fn, message=message)


def save_timeline():
    events_file_path = _get_events_file_path()
    if not events_file_path:
        return
    global _events
    events_to_write = _events
    _events = []
    json_output = {
        'traceEvents': events_to_write,
        'displayTimeUnit': 'ms',
        'otherData': {
            'log_dir': os.path.dirname(os.path.abspath(events_file_path)),
        }
    }
    os.makedirs(os.path.dirname(os.path.abspath(events_file_path)),
                exist_ok=True)
    with open(events_file_path, 'w', encoding='utf-8') as f:
        json.dump(json_output, f)
    del events_to_write


if _get_events_file_path():
    atexit.register(save_timeline)



"""Volume management core."""

import contextlib
import os
from typing import Any, Dict, Generator, List, Optional
import uuid

import filelock

from sky import global_user_state
from sky import models
from sky import provision
from sky import sky_logging
from sky.schemas.api import responses
from sky.utils import common_utils
from sky.utils import registry
from sky.utils import rich_utils
from sky.utils import status_lib
from sky.utils import ux_utils

logger = sky_logging.init_logger(__name__)

# Filelocks for the storage management.
VOLUME_LOCK_PATH = os.path.expanduser('~/.sky/.{volume_name}.lock')
VOLUME_LOCK_TIMEOUT_SECONDS = 20


def volume_refresh() -> None:
    """Refreshes volume status by querying cloud APIs.

    This is called by the background daemon to update volume state.
    It updates status, error messages, and usage information in the database.

    Status transitions:
    - NOT_READY: Volume has errors (e.g., pending due to misconfiguration)
    - IN_USE: Volume is healthy and in use
    - READY: Volume is healthy and not in use
    """
    volumes = global_user_state.get_volumes(is_ephemeral=False)

    # Group volumes by cloud for batch API calls
    cloud_to_configs: Dict[str, List[models.VolumeConfig]] = {}
    volume_name_to_config: Dict[str, models.VolumeConfig] = {}
    for volume in volumes:
        config = volume.get('handle')
        if config is None:
            volume_name = volume.get('name')
            logger.warning(f'Volume {volume_name} has no handle.')
            continue
        cloud = config.cloud
        if cloud not in cloud_to_configs:
            cloud_to_configs[cloud] = []
        cloud_to_configs[cloud].append(config)
        volume_name_to_config[volume.get('name')] = config

    # Check for volume errors (e.g., misconfiguration)
    cloud_to_volume_errors: Dict[str, Dict[str, Optional[str]]] = {}
    for cloud, configs in cloud_to_configs.items():
        try:
            volume_errors = provision.get_all_volumes_errors(cloud, configs)
            cloud_to_volume_errors[cloud] = volume_errors
        except Exception as e:  # pylint: disable=broad-except
            logger.debug(
                f'Failed to get volume errors for volumes on {cloud}: {e}')
            cloud_to_volume_errors[cloud] = {}

    # Get usedby info for all volumes
    cloud_to_used_by_pods: Dict[str, Dict[str, Any]] = {}
    cloud_to_used_by_clusters: Dict[str, Dict[str, Any]] = {}
    cloud_to_failed_volume_names: Dict[str, set] = {}
    for cloud, configs in cloud_to_configs.items():
        try:
            used_by_pods, used_by_clusters, failed_volume_names = (
                provision.get_all_volumes_usedby(cloud, configs))
            cloud_to_used_by_pods[cloud] = used_by_pods
            cloud_to_used_by_clusters[cloud] = used_by_clusters
            cloud_to_failed_volume_names[cloud] = failed_volume_names
        except Exception as e:  # pylint: disable=broad-except
            logger.warning(
                f'Failed to get usedby info for volumes on {cloud}: {e}')
            cloud_to_used_by_pods[cloud] = {}
            cloud_to_used_by_clusters[cloud] = {}
            cloud_to_failed_volume_names[cloud] = {
                config.name for config in configs
            }

    # Update volume statuses in database
    for volume in volumes:
        volume_name = volume.get('name')
        config = volume_name_to_config.get(volume_name)
        if config is None:
            continue

        cloud = config.cloud

        # Skip if usedby fetch failed
        if volume_name in cloud_to_failed_volume_names.get(cloud, set()):
            logger.debug(f'Skipping status update for volume {volume_name} '
                         f'due to failed usedby fetch')
            continue

        # Check for volume errors first
        volume_error = cloud_to_volume_errors.get(cloud, {}).get(volume_name)

        # Get usedby info
        usedby_pods, usedby_clusters = provision.map_all_volumes_usedby(
            cloud,
            cloud_to_used_by_pods.get(cloud, {}),
            cloud_to_used_by_clusters.get(cloud, {}),
            config,
        )

        with _volume_lock(volume_name):
            latest_volume = global_user_state.get_volume_by_name(volume_name)
            if latest_volume is None:
                logger.warning(f'Volume {volume_name} not found.')
                continue

            current_status = latest_volume.get('status')
            current_error = latest_volume.get('error_message')
            current_usedby_pods = latest_volume.get('usedby_pods', [])
            current_usedby_clusters = latest_volume.get('usedby_clusters', [])

            # Determine new status and error_message
            if volume_error:
                new_status = status_lib.VolumeStatus.NOT_READY
                new_error = volume_error
            elif usedby_pods:
                new_status = status_lib.VolumeStatus.IN_USE
                new_error = None
            else:
                new_status = status_lib.VolumeStatus.READY
                new_error = None

            # Update if anything changed
            status_changed = current_status != new_status
            error_changed = current_error != new_error
            usedby_changed = (
                set(current_usedby_pods) != set(usedby_pods) or
                set(current_usedby_clusters) != set(usedby_clusters))

            if status_changed or error_changed or usedby_changed:
                logger.info(f'Update volume {volume_name} status to '
                            f'{new_status.value}'
                            f'{", error: " + new_error if new_error else ""}')
                global_user_state.update_volume_status(
                    volume_name,
                    status=new_status,
                    error_message=new_error,
                    usedby_pods=usedby_pods,
                    usedby_clusters=usedby_clusters)
            volume_config = latest_volume.get('handle')
            if volume_config is None:
                continue
            # For in-cluster volumes created without setting the region
            # explicitly before PR
            # https://github.com/skypilot-org/skypilot/pull/8386, the region
            # will be None. In this case, when the user enables the external
            # kubeconfig, the region will be shown as the default context in
            # the kubeconfig file. We need to refresh the volume config to set
            # the region to the in-cluster context name for these volumes.
            need_refresh, volume_config = provision.refresh_volume_config(
                volume_config.cloud, volume_config)
            if need_refresh:
                global_user_state.update_volume_config(volume_name,
                                                       volume_config)


def volume_list(
    is_ephemeral: Optional[bool] = None,
    refresh: bool = False,
) -> List[responses.VolumeRecord]:
    """Gets volumes from the database.

    Args:
        is_ephemeral: Whether to include ephemeral volumes.
        refresh: If True, refresh volume state from cloud APIs before returning.

    Returns:
        [
            {
                'name': str,
                'type': str,
                'launched_at': int timestamp of creation,
                'cloud': str,
                'region': str,
                'zone': str,
                'size': str,
                'config': Dict[str, Any],
                'name_on_cloud': str,
                'user_hash': str,
                'workspace': str,
                'last_attached_at': int timestamp of last attachment,
                'last_use': last command,
                'status': sky.VolumeStatus,
                'usedby_pods': List[str],
                'usedby_clusters': List[str],
                'usedby_fetch_failed': bool,
                'is_ephemeral': bool,
                'error_message': Optional[str],
            }
        ]
    """
    if refresh:
        volume_refresh()
    with rich_utils.safe_status(ux_utils.spinner_message('Listing volumes')):
        volumes = global_user_state.get_volumes(is_ephemeral=is_ephemeral)
        all_users = global_user_state.get_all_users()
        user_map = {user.id: user.name for user in all_users}

        records = []
        for volume in volumes:
            volume_name = volume.get('name')
            config = volume.get('handle')
            if config is None:
                logger.warning(f'Volume {volume_name} has no handle.')
                continue

            status = volume.get('status')
            record: Dict[str, Any] = {
                'name': volume_name,
                'launched_at': volume.get('launched_at'),
                'user_hash': volume.get('user_hash'),
                'user_name': user_map.get(volume.get('user_hash'), ''),
                'workspace': volume.get('workspace'),
                'last_attached_at': volume.get('last_attached_at'),
                'last_use': volume.get('last_use'),
                'status': status.value if status is not None else '',
                'usedby_pods': volume.get('usedby_pods', []),
                'usedby_clusters': volume.get('usedby_clusters', []),
                'usedby_fetch_failed': False,
                'is_ephemeral': volume.get('is_ephemeral', False),
                'error_message': volume.get('error_message'),
                'type': config.type,
                'cloud': config.cloud,
                'region': config.region,
                'zone': config.zone,
                'size': config.size,
                'config': config.config,
                'name_on_cloud': config.name_on_cloud,
            }
            records.append(responses.VolumeRecord(**record))
        return records


def volume_delete(names: List[str],
                  ignore_not_found: bool = False,
                  purge: bool = False) -> None:
    """Deletes volumes.

    Args:
        names: List of volume names to delete.
        ignore_not_found: If True, ignore volumes that are not found.
        purge: If True, delete the volume from the database even if the
          deletion API fails.

    Raises:
        ValueError: If the volume does not exist
          or is in use or has no handle.
    """
    with rich_utils.safe_status(ux_utils.spinner_message('Deleting volumes')):
        for name in names:
            volume = global_user_state.get_volume_by_name(name)
            if volume is None:
                if ignore_not_found:
                    continue
                raise ValueError(f'Volume {name} not found.')
            config = volume.get('handle')
            if config is None:
                raise ValueError(f'Volume {name} has no handle.')
            cloud = config.cloud
            if not purge:
                usedby_pods, usedby_clusters = provision.get_volume_usedby(
                    cloud, config)
                if usedby_clusters:
                    usedby_clusters_str = ', '.join(usedby_clusters)
                    cluster_str = 'clusters' if len(
                        usedby_clusters) > 1 else 'cluster'
                    raise ValueError(f'Volume {name} is used by {cluster_str}'
                                     f' {usedby_clusters_str}.')
                if usedby_pods:
                    usedby_pods_str = ', '.join(usedby_pods)
                    pod_str = 'pods' if len(usedby_pods) > 1 else 'pod'
                    raise ValueError(
                        f'Volume {name} is used by {pod_str} {usedby_pods_str}.'
                    )
            logger.debug(f'Deleting volume {name} with config {config}')
            with _volume_lock(name):
                try:
                    provision.delete_volume(cloud, config)
                except Exception as e:  # pylint: disable=broad-except
                    if purge:
                        logger.warning(f'Failed to delete volume {name} '
                                       f'on {cloud}: {e}. Purging from '
                                       'database.')
                    else:
                        raise
                global_user_state.delete_volume(name)
        logger.info(f'Deleted volumes: {names}')


def volume_apply(
    name: str,
    volume_type: str,
    cloud: str,
    region: Optional[str],
    zone: Optional[str],
    size: Optional[str],
    config: Dict[str, Any],
    labels: Optional[Dict[str, str]] = None,
    use_existing: Optional[bool] = None,
    is_ephemeral: bool = False,
) -> None:
    """Creates or registers a volume.

    Args:
        name: The name of the volume.
        volume_type: The type of the volume.
        cloud: The cloud of the volume.
        region: The region of the volume.
        zone: The zone of the volume.
        size: The size of the volume.
        config: The configuration of the volume.
        labels: The labels of the volume.
        use_existing: Whether to use an existing volume.
        is_ephemeral: Whether the volume is ephemeral.
    """
    with rich_utils.safe_status(ux_utils.spinner_message('Creating volume')):
        # Reuse the method for cluster name on cloud to
        # generate the storage name on cloud.
        cloud_obj = registry.CLOUD_REGISTRY.from_str(cloud)
        assert cloud_obj is not None
        region, zone = cloud_obj.validate_region_zone(region, zone)
        if use_existing:
            name_on_cloud = name
        else:
            name_uuid = str(uuid.uuid4())[:6]
            name_on_cloud = common_utils.make_cluster_name_on_cloud(
                name, max_length=cloud_obj.max_cluster_name_length())
            name_on_cloud += '-' + name_uuid
        config = models.VolumeConfig(
            name=name,
            type=volume_type,
            cloud=str(cloud_obj),
            region=region,
            zone=zone,
            size=size,
            config=config,
            name_on_cloud=name_on_cloud,
            labels=labels,
        )
        logger.debug(
            f'Creating volume {name} on cloud {cloud} with config {config}')
        with _volume_lock(name):
            current_volume = global_user_state.get_volume_by_name(name)
            if current_volume is not None:
                logger.info(f'Volume {name} already exists.')
                return
            config = provision.apply_volume(cloud, config)
            global_user_state.add_volume(
                name,
                config,
                status_lib.VolumeStatus.READY,
                is_ephemeral,
            )
        logger.info(f'Created volume {name} on cloud {cloud}')


@contextlib.contextmanager
def _volume_lock(volume_name: str) -> Generator[None, None, None]:
    """Context manager for volume lock."""
    try:
        with filelock.FileLock(VOLUME_LOCK_PATH.format(volume_name=volume_name),
                               VOLUME_LOCK_TIMEOUT_SECONDS):
            yield
    except filelock.Timeout as e:
        raise RuntimeError(
            f'Failed to update user due to a timeout '
            f'when trying to acquire the lock at '
            f'{VOLUME_LOCK_PATH.format(volume_name=volume_name)}. '
            'Please try again or manually remove the lock '
            f'file if you believe it is stale.') from e



#!/usr/bin/env python3
# /// script
# dependencies = [
#   "websockets>=14.0",
# ]
# ///
"""Starting a websocket with SkyPilot API server to proxy SSH to a k8s pod.

This script is useful for users who do not have local Kubernetes credentials.
"""
import asyncio
import os
import struct
import sys
import time
from typing import Dict, Optional

import requests
import websockets
from websockets.asyncio.client import ClientConnection
from websockets.asyncio.client import connect

from sky import exceptions
from sky.client import service_account_auth
from sky.server import common as server_common
from sky.server import constants
from sky.server.server import SSHMessageType
from sky.skylet import constants as skylet_constants

BUFFER_SIZE = 2**16  # 64KB
HEARTBEAT_INTERVAL_SECONDS = 10
MAX_UNANSWERED_PINGS = 100
# Timeout for opening the WebSocket connection. The default (10s) can be
# insufficient when many concurrent SSH connections are established under load,
# causing intermittent "timed out during opening handshake" errors.
OPEN_TIMEOUT_SECONDS = 60


async def main(url: str, timestamps_supported: bool, login_url: str) -> None:
    headers = {}
    headers.update(server_common.get_cookie_header_for_url(url))
    headers.update(service_account_auth.get_service_account_headers())
    try:
        async with connect(url,
                           ping_interval=None,
                           open_timeout=OPEN_TIMEOUT_SECONDS,
                           additional_headers=headers) as websocket:
            await run_websocket_proxy(websocket, timestamps_supported)
    except websockets.exceptions.InvalidStatus as e:
        if e.response.status_code == 403:
            print(str(exceptions.ApiServerAuthenticationError(login_url)),
                  file=sys.stderr)
        else:
            print(f'Error ssh into cluster: {e}', file=sys.stderr)
        sys.exit(1)


async def run_websocket_proxy(websocket: ClientConnection,
                              timestamps_supported: bool) -> None:
    if os.isatty(sys.stdin.fileno()):
        # pylint: disable=import-outside-toplevel
        import termios
        import tty
        old_settings = termios.tcgetattr(sys.stdin.fileno())
        tty.setraw(sys.stdin.fileno())
    else:
        old_settings = None

    try:
        loop = asyncio.get_running_loop()
        # Use asyncio.Stream primitives to wrap stdin and stdout, this is to
        # avoid creating a new thread for each read/write operation
        # excessively.
        stdin_reader = asyncio.StreamReader()
        protocol = asyncio.StreamReaderProtocol(stdin_reader)
        await loop.connect_read_pipe(lambda: protocol, sys.stdin)
        transport, protocol = await loop.connect_write_pipe(
            asyncio.streams.FlowControlMixin, sys.stdout)  # type: ignore
        stdout_writer = asyncio.StreamWriter(transport, protocol, None, loop)
        # Dictionary to store last ping time for latency measurement
        last_ping_time_dict: Optional[Dict[int, float]] = None
        if timestamps_supported:
            last_ping_time_dict = {}

        # Use an Event to signal when websocket is closed
        websocket_closed_event = asyncio.Event()
        websocket_lock = asyncio.Lock()

        await asyncio.gather(
            stdin_to_websocket(stdin_reader, websocket, timestamps_supported,
                               websocket_closed_event, websocket_lock),
            websocket_to_stdout(websocket, stdout_writer, timestamps_supported,
                                last_ping_time_dict, websocket_closed_event,
                                websocket_lock),
            latency_monitor(websocket, last_ping_time_dict,
                            websocket_closed_event, websocket_lock),
            return_exceptions=True)
    finally:
        if old_settings:
            termios.tcsetattr(sys.stdin.fileno(), termios.TCSADRAIN,
                              old_settings)


async def latency_monitor(websocket: ClientConnection,
                          last_ping_time_dict: Optional[dict],
                          websocket_closed_event: asyncio.Event,
                          websocket_lock: asyncio.Lock):
    """Periodically send PING messages (type 1) to measure latency."""
    if last_ping_time_dict is None:
        return
    next_id = 0
    while not websocket_closed_event.is_set():
        try:
            await asyncio.sleep(HEARTBEAT_INTERVAL_SECONDS)
            if len(last_ping_time_dict) >= MAX_UNANSWERED_PINGS:
                # We are not getting responses, clear the dictionary so
                # as not to grow unbounded.
                last_ping_time_dict.clear()
            ping_time = time.time()
            next_id += 1
            last_ping_time_dict[next_id] = ping_time
            message_header_bytes = struct.pack('!BI',
                                               SSHMessageType.PINGPONG.value,
                                               next_id)
            try:
                async with websocket_lock:
                    await websocket.send(message_header_bytes)
            except websockets.exceptions.ConnectionClosed as e:
                # Websocket is already closed.
                print(f'Failed to send PING message: {e}', file=sys.stderr)
                break
        except Exception as e:
            print(f'Error in latency_monitor: {e}', file=sys.stderr)
            websocket_closed_event.set()
            raise e


async def stdin_to_websocket(reader: asyncio.StreamReader,
                             websocket: ClientConnection,
                             timestamps_supported: bool,
                             websocket_closed_event: asyncio.Event,
                             websocket_lock: asyncio.Lock):
    try:
        while not websocket_closed_event.is_set():
            # Read at most BUFFER_SIZE bytes, this not affect
            # responsiveness since it will return as soon as
            # there is at least one byte.
            # The BUFFER_SIZE is chosen to be large enough to improve
            # throughput.
            data = await reader.read(BUFFER_SIZE)

            if not data:
                break
            if timestamps_supported:
                # Send message with type 0 to indicate data.
                message_type_bytes = struct.pack(
                    '!B', SSHMessageType.REGULAR_DATA.value)
                data = message_type_bytes + data
            async with websocket_lock:
                await websocket.send(data)

    except Exception as e:  # pylint: disable=broad-except
        print(f'Error in stdin_to_websocket: {e}', file=sys.stderr)
    finally:
        async with websocket_lock:
            await websocket.close()
        websocket_closed_event.set()


async def websocket_to_stdout(websocket: ClientConnection,
                              writer: asyncio.StreamWriter,
                              timestamps_supported: bool,
                              last_ping_time_dict: Optional[dict],
                              websocket_closed_event: asyncio.Event,
                              websocket_lock: asyncio.Lock):
    try:
        while not websocket_closed_event.is_set():
            message = await websocket.recv()
            if (timestamps_supported and len(message) > 0 and
                    last_ping_time_dict is not None):
                message_type = struct.unpack('!B', message[:1])[0]
                if message_type == SSHMessageType.REGULAR_DATA.value:
                    # Regular data - strip type byte and write to stdout
                    message = message[1:]
                elif message_type == SSHMessageType.PINGPONG.value:
                    # PONG response - calculate latency and send measurement
                    if not len(message) == struct.calcsize('!BI'):
                        raise ValueError(
                            f'Invalid PONG message length: {len(message)}')
                    pong_id = struct.unpack('!I', message[1:5])[0]
                    pong_time = time.time()

                    ping_time = last_ping_time_dict.pop(pong_id, None)

                    if ping_time is None:
                        continue

                    latency_seconds = pong_time - ping_time
                    latency_ms = int(latency_seconds * 1000)

                    # Send latency measurement (type 2)
                    message_type_bytes = struct.pack(
                        '!B', SSHMessageType.LATENCY_MEASUREMENT.value)
                    latency_bytes = struct.pack('!Q', latency_ms)
                    message = message_type_bytes + latency_bytes
                    # Send to server.
                    async with websocket_lock:
                        await websocket.send(message)
                    continue
            # No timestamps support, write directly
            writer.write(message)
            await writer.drain()
    except websockets.exceptions.ConnectionClosed:
        print('WebSocket connection closed', file=sys.stderr)
    except Exception as e:  # pylint: disable=broad-except
        print(f'Error in websocket_to_stdout: {e}', file=sys.stderr)
        raise e
    finally:
        async with websocket_lock:
            await websocket.close()
        websocket_closed_event.set()


if __name__ == '__main__':
    server_url = sys.argv[1].strip('/')

    disable_latency_measurement = os.environ.get(
        skylet_constants.SSH_DISABLE_LATENCY_MEASUREMENT_ENV_VAR, '0') == '1'
    if disable_latency_measurement:
        timestamps_are_supported = False
    else:
        # TODO(aylei): remove the separate /api/health call and use the header
        # during websocket handshake to determine the server version.
        health_url = f'{server_url}/api/health'
        cookie_hdr = server_common.get_cookie_header_for_url(health_url)
        health_response = requests.get(health_url, headers=cookie_hdr)
        health_data = health_response.json()
        timestamps_are_supported = int(health_data.get('api_version', 0)) > 21

    # Capture the original API server URL for login hint if authentication
    # is required.
    _login_url = server_url
    server_proto, server_fqdn = server_url.split('://')
    websocket_proto = 'ws'
    if server_proto == 'https':
        websocket_proto = 'wss'
    server_url = f'{websocket_proto}://{server_fqdn}'

    client_version_str = (f'&client_version={constants.API_VERSION}'
                          if timestamps_are_supported else '')

    # For backwards compatibility, fallback to kubernetes-pod-ssh-proxy if
    # no endpoint is provided.
    endpoint = sys.argv[3] if len(sys.argv) > 3 else 'kubernetes-pod-ssh-proxy'
    # Worker index for Slurm.
    worker_idx = sys.argv[4] if len(sys.argv) > 4 else '0'
    websocket_url = (f'{server_url}/{endpoint}'
                     f'?cluster_name={sys.argv[2]}'
                     f'&worker={worker_idx}'
                     f'{client_version_str}')
    asyncio.run(main(websocket_url, timestamps_are_supported, _login_url))



"""Prime Intellect instance provisioning."""
import time
from typing import Any, Dict, List, Optional, Tuple

from sky import exceptions
from sky import sky_logging
from sky.provision import common
from sky.provision.primeintellect import utils
from sky.utils import common_utils
from sky.utils import status_lib
from sky.utils import ux_utils

# The maximum number of times to poll for the status of an operation.
POLL_INTERVAL = 5
MAX_POLLS = 60 // POLL_INTERVAL
# Terminating instances can take several minutes, so we increase the timeout
MAX_POLLS_FOR_UP_OR_TERMINATE = MAX_POLLS * 16

# status filters
# PROVISIONING, PENDING, ACTIVE, STOPPED, ERROR, DELETING, TERMINATED

logger = sky_logging.init_logger(__name__)

# SSH connection readiness polling constants
SSH_CONN_MAX_RETRIES = 6
SSH_CONN_RETRY_INTERVAL_SECONDS = 10


def _filter_instances(cluster_name_on_cloud: str,
                      status_filters: Optional[List[str]]) -> Dict[str, Any]:
    client = utils.PrimeIntellectAPIClient()
    instances = client.list_instances()
    # TODO: verify names are we using it?
    possible_names = [
        f'{cluster_name_on_cloud}-head',
        f'{cluster_name_on_cloud}-worker',
    ]

    filtered_instances = {}
    for instance in instances:
        instance_id = instance['id']
        if (status_filters is not None and
                instance['status'] not in status_filters):
            continue
        instance_name = instance.get('name')
        if instance_name and instance_name in possible_names:
            filtered_instances[instance_id] = instance
    return filtered_instances


def _get_instance_info(instance_id: str) -> Dict[str, Any]:
    client = utils.PrimeIntellectAPIClient()
    return client.get_instance_details(instance_id)


def _get_head_instance_id(instances: Dict[str, Any]) -> Optional[str]:
    head_instance_id = None
    for inst_id, inst in instances.items():
        if inst['name'].endswith('-head'):
            head_instance_id = inst_id
            break
    return head_instance_id


# Helper is available as utils.parse_ssh_connection.


def run_instances(region: str, cluster_name: str, cluster_name_on_cloud: str,
                  config: common.ProvisionConfig) -> common.ProvisionRecord:
    """Runs instances for the given cluster."""
    del cluster_name  # unused
    pending_status = [
        'PROVISIONING',
        'PENDING',
    ]
    newly_started_instances = _filter_instances(cluster_name_on_cloud,
                                                pending_status)
    client = utils.PrimeIntellectAPIClient()

    while True:
        instances = _filter_instances(cluster_name_on_cloud, pending_status)
        if not instances:
            break
        instance_statuses = [
            instance['status'] for instance in instances.values()
        ]
        logger.info(f'Waiting for {len(instances)} instances to be ready: '
                    f'{instance_statuses}')
        time.sleep(POLL_INTERVAL)

    exist_instances = _filter_instances(cluster_name_on_cloud,
                                        status_filters=pending_status)
    if len(exist_instances) > config.count:
        raise RuntimeError(
            f'Cluster {cluster_name_on_cloud} already has '
            f'{len(exist_instances)} nodes, but {config.count} are required.')

    exist_instances = _filter_instances(cluster_name_on_cloud,
                                        status_filters=['ACTIVE'])
    head_instance_id = _get_head_instance_id(exist_instances)
    to_start_count = config.count - len(exist_instances)
    if to_start_count < 0:
        raise RuntimeError(
            f'Cluster {cluster_name_on_cloud} already has '
            f'{len(exist_instances)} nodes, but {config.count} are required.')
    if to_start_count == 0:
        if head_instance_id is None:
            head_instance_id = list(exist_instances.keys())[0]
            # TODO: implement rename pod
            # client.rename(
            #     instance_id=head_instance_id,
            #     name=f'{cluster_name_on_cloud}-head',
            # )
        assert head_instance_id is not None, (
            'head_instance_id should not be None')
        logger.info(f'Cluster {cluster_name_on_cloud} already has '
                    f'{len(exist_instances)} nodes, no need to start more.')
        return common.ProvisionRecord(
            provider_name='primeintellect',
            cluster_name=cluster_name_on_cloud,
            region=region,
            zone=config.provider_config['zones'],
            head_instance_id=head_instance_id,
            resumed_instance_ids=list(newly_started_instances.keys()),
            created_instance_ids=[],
        )

    created_instance_ids = []
    for _ in range(to_start_count):
        node_type = 'head' if head_instance_id is None else 'worker'
        try:
            # Extract vCPUs and memory from instance type
            # Format: provider__gpu_prefix_base_type__vcpus__memory[_SPOT]
            instance_type = config.node_config['InstanceType']
            disk_size = config.node_config.get('DiskSize')
            vcpus = -1
            memory = -1
            try:
                # Split by '__'
                parts = instance_type.split('__')

                # Format: provider__gpu_info__vcpus__memory[_SPOT]
                # For: primecompute__8xH100_80GB__104__752_SPOT
                # parts[0] = primecompute, parts[1] = 8xH100_80GB,
                # parts[2] = 104, parts[3] = 752, parts[4] = SPOT
                if len(parts) >= 4:
                    vcpu_str = parts[2]
                    memory_str = parts[3]
                    vcpus = int(vcpu_str)
                    memory = int(memory_str)
            except (ValueError, IndexError) as e:
                # If parsing fails, try to get from catalog
                logger.warning(
                    f'Failed to parse vCPUs/memory from instance type '
                    f'{instance_type}: {e}')

            params = {
                'name': f'{cluster_name_on_cloud}-{node_type}',
                'instance_type': config.node_config['InstanceType'],
                'region': region,
                'availability_zone': config.provider_config['zones'],
                'disk_size': disk_size,
                'vcpus': vcpus,
                'memory': memory,
            }

            response = client.launch(**params)
            instance_id = response['id']
        except utils.PrimeintellectResourcesUnavailableError as e:
            # Resource unavailability error - provide specific message
            instance_type = config.node_config['InstanceType']
            region_str = (f' in region {region}'
                          if region != 'PLACEHOLDER' else '')
            error_msg = (
                f'Resources are currently unavailable on Prime Intellect. '
                f'No {instance_type} instances are available{region_str}. '
                f'Please try again later or consider using a different '
                f'instance type or region. Details: {str(e)}')
            logger.warning(f'Resource unavailability error: {e}')
            with ux_utils.print_exception_no_traceback():
                raise exceptions.ResourcesUnavailableError(error_msg) from e
        except utils.PrimeintellectAPIError as e:
            # Other API errors - provide specific message
            instance_type = config.node_config['InstanceType']
            region_str = (f' in region {region}'
                          if region != 'PLACEHOLDER' else '')
            error_msg = (f'Failed to launch {instance_type} instance on Prime '
                         f'Intellect{region_str}. Details: {str(e)}')
            logger.warning(f'API error during instance launch: {e}')
            with ux_utils.print_exception_no_traceback():
                raise exceptions.ResourcesUnavailableError(error_msg) from e
        except Exception as e:  # pylint: disable=broad-except
            # Generic error handling for unexpected errors
            instance_type = config.node_config['InstanceType']
            region_str = (f' in region {region}'
                          if region != 'PLACEHOLDER' else '')
            error_msg = (
                f'Unexpected error while launching {instance_type} instance '
                f'on Prime Intellect{region_str}. Details: '
                f'{common_utils.format_exception(e, use_bracket=False)}')
            logger.warning(f'Unexpected error during instance launch: {e}')
            with ux_utils.print_exception_no_traceback():
                raise exceptions.ResourcesUnavailableError(error_msg) from e
        logger.info(f'Launched instance {instance_id}.')
        created_instance_ids.append(instance_id)
        if head_instance_id is None:
            head_instance_id = instance_id

    # Wait for instances to be ready.
    for _ in range(MAX_POLLS_FOR_UP_OR_TERMINATE):
        instances = _filter_instances(cluster_name_on_cloud, ['ACTIVE'])
        logger.info('Waiting for instances to be ready: '
                    f'({len(instances)}/{config.count}).')
        if len(instances) == config.count:
            break

        time.sleep(POLL_INTERVAL)
    else:
        # Failed to launch config.count of instances after max retries
        # Provide more specific error message
        instance_type = config.node_config['InstanceType']
        region_str = (f' in region {region}' if region != 'PLACEHOLDER' else '')
        active_instances = len(
            _filter_instances(cluster_name_on_cloud, ['ACTIVE']))
        error_msg = (
            f'Timed out waiting for {instance_type} instances to become '
            f'ready on Prime Intellect{region_str}. Only {active_instances} '
            f'out of {config.count} instances became active. This may '
            f'indicate capacity issues or slow provisioning. Please try '
            f'again later or consider using a different instance type or '
            f'region.')
        logger.warning(error_msg)
        with ux_utils.print_exception_no_traceback():
            raise exceptions.ResourcesUnavailableError(error_msg)
    assert head_instance_id is not None, 'head_instance_id should not be None'
    return common.ProvisionRecord(
        provider_name='primeintellect',
        cluster_name=cluster_name_on_cloud,
        region=region,
        zone=config.provider_config['zones'],
        head_instance_id=head_instance_id,
        resumed_instance_ids=[],
        created_instance_ids=created_instance_ids,
    )


def wait_instances(region: str, cluster_name_on_cloud: str,
                   state: Optional[status_lib.ClusterStatus]) -> None:
    del region, cluster_name_on_cloud, state


def stop_instances(
    cluster_name_on_cloud: str,
    provider_config: Optional[Dict[str, Any]] = None,
    worker_only: bool = False,
) -> None:
    raise NotImplementedError()


def terminate_instances(
    cluster_name_on_cloud: str,
    provider_config: Optional[Dict[str, Any]] = None,
    worker_only: bool = False,
) -> None:
    """See sky/provision/__init__.py"""
    del provider_config  # unused
    client = utils.PrimeIntellectAPIClient()
    instances = _filter_instances(cluster_name_on_cloud, None)

    # Log if no instances found
    if not instances:
        logger.info(f'No instances found for cluster {cluster_name_on_cloud}')
        return

    # Filter out already terminated instances
    non_terminated_instances = {
        inst_id: inst
        for inst_id, inst in instances.items()
        if inst['status'] not in ['TERMINATED', 'DELETING']
    }

    if not non_terminated_instances:
        logger.info(
            f'All instances for cluster {cluster_name_on_cloud} are already '
            f'terminated or being deleted')
        return

    # Log what we're about to terminate
    instance_names = [
        inst['name'] for inst in non_terminated_instances.values()
    ]
    logger.info(
        f'Terminating {len(non_terminated_instances)} instances for cluster '
        f'{cluster_name_on_cloud}: {instance_names}')

    # Terminate each instance
    terminated_instances = []
    for inst_id, inst in non_terminated_instances.items():
        status = inst['status']
        logger.debug(f'Terminating instance {inst_id} (status: {status})')
        if worker_only and inst['name'].endswith('-head'):
            continue
        try:
            client.remove(inst_id)
            terminated_instances.append(inst_id)
            name = inst['name']
            logger.info(
                f'Successfully initiated termination of instance {inst_id} '
                f'({name})')
        except Exception as e:  # pylint: disable=broad-except
            with ux_utils.print_exception_no_traceback():
                raise RuntimeError(
                    f'Failed to terminate instance {inst_id}: '
                    f'{common_utils.format_exception(e, use_bracket=False)}'
                ) from e

    # Wait for instances to be terminated
    if not terminated_instances:
        logger.info(
            'No instances were terminated (worker_only=True and only head '
            'node found)')
        return

    logger.info(f'Waiting for {len(terminated_instances)} instances to be '
                f'terminated...')
    for _ in range(MAX_POLLS_FOR_UP_OR_TERMINATE):
        remaining_instances = _filter_instances(cluster_name_on_cloud, None)

        # Check if all terminated instances are gone
        still_exist = [
            inst_id for inst_id in terminated_instances
            if inst_id in remaining_instances
        ]
        if not still_exist:
            logger.info('All instances have been successfully terminated')
            break

        # Log status of remaining instances
        remaining_statuses = [(inst_id, remaining_instances[inst_id]['status'])
                              for inst_id in still_exist]
        logger.info(
            f'Waiting for termination... {len(still_exist)} instances still '
            f'exist: {remaining_statuses}')
        time.sleep(POLL_INTERVAL)
    else:
        # Timeout reached
        remaining_instances = _filter_instances(cluster_name_on_cloud, None)
        still_exist = [
            inst_id for inst_id in terminated_instances
            if inst_id in remaining_instances
        ]
        if still_exist:
            logger.warning(
                f'Timeout reached. {len(still_exist)} instances may still be '
                f'terminating: {still_exist}')
        else:
            logger.info('All instances have been successfully terminated')


def get_cluster_info(
        region: str,
        cluster_name_on_cloud: str,
        provider_config: Optional[Dict[str, Any]] = None) -> common.ClusterInfo:
    del region  # unused
    running_instances = _filter_instances(cluster_name_on_cloud, ['ACTIVE'])
    instances: Dict[str, List[common.InstanceInfo]] = {}
    head_instance_id = None
    head_ssh_user = None
    for instance_id, instance in running_instances.items():
        retry_count = 0
        max_retries = SSH_CONN_MAX_RETRIES
        while (instance.get('sshConnection') is None and
               retry_count < max_retries):
            name = instance.get('name')
            print(f'SSH connection to {name} is not ready, waiting '
                  f'{SSH_CONN_RETRY_INTERVAL_SECONDS} seconds... '
                  f'(attempt {retry_count + 1}/{max_retries})')
            time.sleep(SSH_CONN_RETRY_INTERVAL_SECONDS)
            retry_count += 1
            running_instances[instance_id] = _get_instance_info(instance_id)

        if instance.get('sshConnection') is not None:
            print('SSH connection is ready!')
        else:
            raise Exception(
                f'Failed to establish SSH connection after {max_retries} '
                f'attempts')

        assert instance.get(
            'sshConnection'), 'sshConnection cannot be null anymore'

        ssh_connection = instance['sshConnection']
        _, ssh_port = utils.parse_ssh_connection(ssh_connection)

        external_ip = instance['ip']
        if isinstance(external_ip, list):
            external_ip = external_ip[0]

        instances[instance_id] = [
            common.InstanceInfo(
                instance_id=instance_id,
                internal_ip='NOT_SUPPORTED',
                external_ip=external_ip,
                ssh_port=ssh_port,
                tags={'provider': instance['providerType']},
                node_name=instance_id,
            )
        ]
        if instance['name'].endswith('-head'):
            head_instance_id = instance_id
            parsed_user_for_user, _ = utils.parse_ssh_connection(ssh_connection)
            head_ssh_user = parsed_user_for_user or 'ubuntu'

    return common.ClusterInfo(
        instances=instances,
        head_instance_id=head_instance_id,
        provider_name='primeintellect',
        provider_config=provider_config,
        ssh_user=head_ssh_user,
    )


def query_instances(
    cluster_name_on_cloud: str,
    provider_config: Optional[Dict[str, Any]] = None,
    non_terminated_only: bool = True,
) -> Dict[str, Tuple[Optional['status_lib.ClusterStatus'], Optional[str]]]:
    """See sky/provision/__init__.py"""
    assert provider_config is not None, (cluster_name_on_cloud, provider_config)
    instances = _filter_instances(cluster_name_on_cloud, None)

    status_map = {
        'PENDING': status_lib.ClusterStatus.INIT,
        'ERROR': status_lib.ClusterStatus.INIT,
        'ACTIVE': status_lib.ClusterStatus.UP,
        'STOPPED': status_lib.ClusterStatus.STOPPED,
        'DELETING': None,  # Being deleted - should be filtered out
        'TERMINATED': None,  # Already terminated - should be filtered out
    }
    statuses: Dict[str, Tuple[Optional[status_lib.ClusterStatus],
                              Optional[str]]] = {}
    for inst_id, inst in instances.items():
        status = status_map[inst['status']]
        if non_terminated_only and status is None:
            continue
        statuses[inst_id] = (status, None)
    return statuses


def cleanup_ports(
    cluster_name_on_cloud: str,
    ports: List[str],
    provider_config: Optional[Dict[str, Any]] = None,
) -> None:
    del cluster_name_on_cloud, ports, provider_config  # Unused.
